unit I2LDataTypes;

interface

uses SysUtils,Classes,DSupport, HugeArry, XPLDataTypes;


type TCodeArray = array of byte;

type TI2LProgram = record
 Code: TCodeArray;
 BaseAddress: integer;
 Loaded: boolean;
 end;

{Create an object that combines the Heap and stack so they
appear to be in the same memory space while they are also
accessible separately}

{
Memory Map

1. Base Memory = 0
2. Program Data
3. End of Progam = LocalLow
4. Local Data
5. Start of Heap = HeapLow
6. Top Of Heap = HeapPtr
7. Total size of Heap = HeapSize

}


type TI2LDataBlock = class(TObject)
 private
  FHeapLimit: integer;
  procedure SetHeapSize(const Value: integer);
  function GetDataByte(Inx: integer): byte;
  procedure SetDataByte(Inx: integer; const Value: byte);
  function GetCodeSize: integer;
  function GetTotalSize: integer;
  function GetHeapSize: integer;
 protected
 public
  ProgData: TI2LProgram;	{ Program Storage array}
  ProgCntr: integer;
  LocalLow: integer;
  HeapLow: integer;
  HeapPtr: integer;
  LocalPtr: integer;
  constructor Create;
  procedure InitializeHeap(HeapSize,HeapLimit,LocalSize: integer);
  function ExpandMemory(NewSize: integer): boolean;
  function ReserveMemory(Size: integer): boolean;
  procedure ReleaseMemory;
  property HeapSize: integer read GetHeapSize write SetHeapSize default 32768;
  property HeapLimit: integer read FHeapLimit write FHeapLimit;

  function AddLocal(Value: TI2LWord): TI2LWord;
  procedure InsertLocal(Inx, Value: TI2LWord);

  function RangeCheck(Inx: integer): boolean;
  function GetOpcode: byte;
  function GetProgramByte: byte;
  function GetProgramDouble: double;
  function GetProgramWord: TI2LWord;
  function GetDataPDoubleArray(Inx: integer): PDoubleArray;
  function GetDataDouble(Inx: integer): double;
  function GetDataString(Inx: TI2lWord; MSBTerm: boolean): string;
  function GetDataWord(Inx: integer): TI2LWord;
  procedure SetDataString(S: string; Inx: integer);
  procedure SetDataWord(Inx: integer; const Value: TI2LWord);
  procedure SetDataDouble(Inx: integer; const Value: double);
  property DataByte[Inx: integer]: byte read GetDataByte write SetDataByte;

  property CodeSize: integer read GetCodeSize;
  property TotalSize: integer read GetTotalSize;
 end;


{Memory manager that simulates DOS by allocating memory as pointers paragraphs}


type TSegData = array of byte;


type TSegArray = array of TSegData;


type TMemoryManager = class(TObject)
 private
  function GetAllocSize: integer;
 protected
 public
  SegArray: TSegArray;
  function Allocate(Size: integer): integer;
  function Reallocate(Inx, Size: integer): integer;
  function Deallocate(Inx: integer): integer;
  procedure Clear;
  property AllocSize: integer read GetAllocSize;
 end;


implementation


{ TIT2LDataBlock }

constructor TI2LDataBlock.Create;
begin
inherited Create;
FHeapLimit:=32767;
end;

{-----------------------------------------------------------------------------}

function TI2LDataBlock.GetProgramByte: byte;
begin
Result:=ProgData.Code[ProgCntr];
Inc(ProgCntr);
end;


function TI2LDataBlock.GetOpcode: byte;
begin
{This can be slow because it is only used by the debugger}
if (ProgCntr<0) or (ProgCntr>High(ProgData.Code)) then Result:=0
else Result:=ProgData.Code[ProgCntr];
end;



function TI2LDataBlock.GetProgramWord: TI2LWord;
begin
Result:=PI2LWord(@ProgData.Code[ProgCntr])^;
Inc(ProgCntr,SizeOf(TI2LWord));
end;



function TI2LDataBlock.GetProgramDouble: double;
begin
Result:=PDouble(@ProgData.Code[ProgCntr])^;
Inc(ProgCntr,SizeOf(double));
end;


{-----------------------------------------------------------------------------}


function TI2LDataBlock.RangeCheck(Inx: integer): boolean;
begin
Result:=(Inx>-1) and (Inx<Length(ProgData.Code));
end;


function TI2LDataBlock.GetDataByte(Inx: integer): byte;
begin
Result:=ProgData.Code[Inx];
end;





procedure TI2LDataBlock.SetDataByte(Inx: integer; const Value: byte);
begin
ProgData.Code[Inx]:=Value;
end;



function TI2LDataBlock.GetDataWord(Inx: integer): TI2LWord;
begin
Result:=PI2LWord(@ProgData.Code[Inx])^;
end;



function TI2LDataBlock.GetDataString(Inx: TI2lWord; MSBTerm: boolean): string;
var I: integer;
var B: byte;
begin
Result:='';
if MSBTerm then
	begin
	for I:=Inx to TotalSize-1 do
		begin
		B:=DataByte[I];
		Result:=Result + Char(B and $7F);
		if B>$7F then break;
		end;
	end
else
	begin
	for I:=Inx to TotalSize-1 do
		begin
		B:=DataByte[I];
		if B=0 then break;
		Result:=Result + char(B);
		end;
	end;
end;



procedure TI2LDataBlock.SetDataString(S: string; Inx: integer);
var I: integer;
begin
for I:=1 to Length(S) do DataByte[Inx+I-1]:=byte(S[I]);
if Length(S)<1 then DataByte[Inx]:=$A0
else DataByte[Inx+Length(S)-1]:=DataByte[Inx+Length(S)-1] or $80;
end;



function TI2LDataBlock.GetDataDouble(Inx: integer): double;
begin
Result:=Pdouble(@ProgData.Code[Inx])^;
end;




function TI2LDataBlock.GetDataPDoubleArray(Inx: integer): PDoubleArray;
begin
Result:=PDoubleArray(@ProgData.Code[Inx]);
end;


procedure TI2LDataBlock.SetDataDouble(Inx: integer; const Value: double);
begin
Pdouble(@ProgData.Code[Inx])^:=Value;
end;




procedure TI2LDataBlock.SetDataWord(Inx: integer; const Value: TI2LWord);
begin
PI2LWord(@ProgData.Code[Inx])^:=Value;
end;



function TI2LDataBlock.GetCodeSize: integer;
begin
Result:=Length(ProgData.Code);
end;



function TI2LDataBlock.GetTotalSize: integer;
begin
Result:=HeapSize+CodeSize;
end;




procedure TI2LDataBlock.InitializeHeap(HeapSize,HeapLimit,LocalSize: integer);
var I: integer;
begin
LocalPtr:=0;
LocalLow:=Length(ProgData.Code);
SetLength(ProgData.Code,LocalLow+LocalSize);
HeapLow:=Length(ProgData.Code);
HeapPtr:=HeapLow;
SetLength(ProgData.Code,HeapLow+HeapSize);
for I:=HeapLow to HeapLow+HeapSize-1 do ProgData.Code[I]:=0;
FHeapLimit:=HeapLimit;
end;



function TI2LDataBlock.ExpandMemory(NewSize: integer): boolean;
begin
Result:=False;
if NewSize>Length(ProgData.Code) then
	begin
	try
	SetLength(ProgData.Code,NewSize);
	except exit; end;
	end;
Result:=True;
end;



function TI2LDataBlock.ReserveMemory(Size: integer): boolean;
var NewSize: integer;
begin
Result:=False;
if Size<0 then exit;
NewSize:=HeapPtr + Size;
if NewSize>=HeapLimit then exit
else
	begin
	Inc(HeapPtr,Size);
	if not ExpandMemory(NewSize) then exit;
	end;
Result:=True;
end;



procedure TI2LDataBlock.ReleaseMemory;
begin
SetLength(ProgData.Code,0);
end;



function TI2LDataBlock.GetHeapSize: integer;
begin
Result:=Length(ProgData.Code)-HeapLow;
if Result<0 then Result:=0;
end;



procedure TI2LDataBlock.SetHeapSize(const Value: integer);
begin
if HeapSize<>Value then
	begin
	SetLength(ProgData.Code,HeapLow+Value);
	end
end;



procedure TI2LDataBlock.InsertLocal(Inx,Value: TI2LWord);
begin
if (Inx<0) or (LocalLow+Inx>=HeapLow) then raise Exception.Create('Outside Local Storage');
SetDataWord(LocalLow+Inx,Value);
end;



function TI2LDataBlock.AddLocal(Value: TI2LWord): TI2LWord;
{Add value to local storage - between program and heap}
begin
if (LocalLow+LocalPtr)>=HeapLow then raise Exception.Create('Local Storage Exceeded');
Result:=LocalPtr;
SetDataWord(LocalLow+LocalPtr,Value);
Inc(LocalPtr,SizeOf(TI2LWord));
end;



{-----------------------------------------------------------------------------}



{ TMemoryManager }

function TMemoryManager.Allocate(Size: integer): integer;
begin
SetLength(SegArray,Length(SegArray)+1);
Result:=High(SegArray);
SetLength(SegArray[Result],Size);
end;



function TMemoryManager.Reallocate(Inx,Size: integer): integer;
begin
Result:=Inx;
SetLength(SegArray[Inx],Size);
end;



procedure TMemoryManager.Clear;
var I: integer;
begin
for I:=0 to High(SegArray) do SetLength(SegArray[I],0);
SetLength(SegArray,0);
end;



function TMemoryManager.Deallocate(Inx: integer): integer;
begin
if (Inx<0) or (Inx>High(SegArray)) then raise Exception.Create('Invalide Segment');
{We release the memory but don't delete the segment}
{Othewise the segments will move and invalidate the paragraphs}
SetLength(SegArray[Inx],0);
end;


function TMemoryManager.GetAllocSize: integer;
var I: integer;
begin
Result:=0;
for I:=0 to High(SegArray) do
	begin
	Inc(Result);
	Inc(Result,Length(SegArray[I]));
	end;
end;



end.
