unit I2LDisasm;

interface

uses SysUtils,Classes,I2LDataTypes,DSupport, XPLDataTypes;


{Define Operand types}

type TOperand = (otNone,otGlobal,otLevel,otOffset,otAddress,otByteHex,otByteDec,otWordHex,otWordDec,otFloat,otData);
type TOpArray = array of TOperand;

{Structure used to dissect and analyze an opcode}

type TDisasmInfo = record
 Name: string;
 Address: TI2LWord;
 Opcode: byte;
 Operands: TOpArray;
 ByteValue: byte;
 WordValue: TI2LWord;
 FloatNum: array [0..7] of byte;
 Data: array of byte;
 DisasmStr: string;
 LineCnt,LineNum: integer;
 end;

{Define an array of dissected opcodes which constitutes the whole disassembly}
{Using object to optimise for speed}

type TDisasmList = class(TObject)
 private
  FArray: array of TDisasmInfo;
  FIndex: integer;
  procedure Put(Index: integer; Item: TDisasmInfo);
  function Get(Index: integer): TDisasmInfo;
 protected
 public
  procedure Clear;
  procedure Add(Info: TDisasmInfo);
  property Count: integer read FIndex;
  property Value[Index: integer]: TDisasmInfo read Get write Put; default;
 end;



procedure DisAssembleI2L(I2L: TCodeArray); overload;
procedure DisAssembleI2L(I2L: TCodeArray; var Disasm: string); overload;
procedure DisAsmI2LWithSource(I2LBin: TI2LProgram; Source: string; var Disasm: string);

function GetDataDump(I2L: TCodeArray; Parse: boolean): string;
function PCToDisasmLine(PC: integer): integer;


implementation

uses XPLCompiler, I2LInterp;


type TFlagArray = array of boolean;	{Array to determine if individual bytes are code or data}

var FlagArray: TFlagArray;		{Actual global array that determines whether a byte is code/data}




{ TDisasmList }

procedure TDisasmList.Add(Info: TDisasmInfo);
const Expand = 64;
begin
if FIndex>High(FArray) then SetLength(FArray,Length(FArray)+Expand);
FArray[FIndex]:=Info;
Inc(FIndex);
end;



procedure TDisasmList.Clear;
begin
SetLength(FArray,0);
FIndex:=0;
end;

function TDisasmList.Get(Index: integer): TDisasmInfo;
begin
Result:=FArray[Index];
end;

procedure TDisasmList.Put(Index: integer; Item: TDisasmInfo);
begin
FArray[Index]:=Item;
end;


{-------- Routines to pre-parse code and find data vs code -------------------}

function GetOpcodeBytes(Opcode: byte): integer;
begin
case Opcode of
 $00: Result:=1;		{1,EXIT INTERPRETER}
 $01: Result:=2+IntBytes;	{4,LOAD A VARIABLE}
 $02: Result:=1;		{1,INDEXED LOAD A BYTE}
 $03: Result:=2+IntBytes;	{4,STORE INTO A VARIABLE}
 $04: Result:=1;		{1,INDEXED STORE TO A BYTE}
 $05: Result:=2+IntBytes;	{4,CALL AN I2L PROCEDURE}
 $06: Result:=1;		{1,RETURN FROM I2L PROCEDURE}
 $07: Result:=1+IntBytes;	{3,JUMP TO I2L CODE}
 $08: Result:=1+IntBytes;	{3,JUMP IF FALSE}
 $09: Result:=1+IntBytes;	{3,INCREMENT HEAP PTR BY ARG}
 $0A: Result:=2;		{2,GET PROC ARGUMENTS}
 $0B: Result:=1+IntBytes;	{3,IMMEDIATE LOAD OF ARG}
 $0C: Result:=2;		{2,CALL AN INTRINSIC}
 $0D: Result:=1;		{1,ADD}
 $0E: Result:=1;		{1,SUBTRACT}
 $0F: Result:=1;		{1,MULTIPLY}
 $10: Result:=1;		{1,DIVIDE}
 $11: Result:=1;		{1,UNARY MINUS}
 $12: Result:=1;		{1,TEST FOR EQUAL}
 $13: Result:=1;		{1,TEST FOR NOT EQUAL}
 $14: Result:=1;		{1,TEST FOR >=}
 $15: Result:=1;		{1,TEST FOR >}
 $16: Result:=1;		{1,TEST FOR <=}
 $17: Result:=1;		{1,TEST FOR <}
 $18: Result:=1+IntBytes;	{3,'FOR' LOOP CONTROL}
 $19: Result:=2+IntBytes;	{4,INCREMENT AND PUSH}
 $1A: Result:=1;		{1,OR}
 $1B: Result:=1;		{1,AND}
 $1C: Result:=1;		{1,1'S COMPLEMENT}
 $1D: Result:=1;		{1,EOR}
 $1E: Result:=1;		{1,NOS+TOS*2 (DBA)}
 $1F: Result:=1;		{1,INDIRECT SAVE (STD)}
 $20: Result:=1;		{1,INDIRECT GET (DBX)}
 $21: Result:=2+IntBytes;	{4,ADDRESS OF VARIABLE}
 $22: Result:=1;		{1,INDIRECT GET}
 $23: Result:=1+IntBytes;	{3,ABSOLUTE GET}
 $24: Result:=2;		{2,SHORT IMMEDIATE}
 $25: Result:=1+IntBytes;	{3,CASE JUMP}
 $26: Result:=1+IntBytes;	{3,OPTIMIZED PROC CALL}
 $27: Result:=1;		{1,OPTIMIZED RETURN}
 $28: Result:=1;		{1,DROP TOS}
 $29: Result:=1+IntBytes;	{3,CALL EXTERNAL SUBROUTINE}
 $2A: Result:=2+IntBytes;	{4,FLOATING LOAD}
 $2B: Result:=2+IntBytes;	{4,FLOATING STORE}
 $2C: Result:=9;		{9,FLOATING LOAD IMMED}
 $2D: Result:=1;		{1,FLOATING ADDITION}
 $2E: Result:=1;		{1,FLOATING SUBTRACT}
 $2F: Result:=1;		{1,FLOATING MULTIPLY}
 $30: Result:=1;		{1,FLOATING DIVIDE}
 $31: Result:=1;		{1,FLOATING NEGATE}
 $32: Result:=1;		{1,TEST FLOATING FOR =}
 $33: Result:=1;		{1,TEST FLOATING FOR #}
 $34: Result:=1;		{1,TEST FLOATING FOR >=}
 $35: Result:=1;		{1,TEST FLOATING FOR >}
 $36: Result:=1;		{1,TEST FLOATING FOR <=}
 $37: Result:=1;		{1,TEST FLOATING FOR <}
 $38: Result:=1;		{1,TOS:= 5*TOS +NOS}
 $39: Result:=1;		{1,PUSH @(5*TOS+NOS)}
 $3A: Result:=1;		{1,PUSH(5)@(TOS(2)+NOS(5))}
 $3B: Result:=1;		{1,PULL(5) @(NOS(2))}
 $3C: Result:=2+IntBytes;	{4,Push address of real}
 $3D: Result:=1;		{1,Arithmetic Shift Right}
 $3E: Result:=1;		{1,SHIFT LEFT}
 $3F: Result:=1;		{1,SHIFT RIGHT}
 $40: Result:=1;		{1,LOAD SEGMENT INTEGER}
 $41: Result:=1;		{1,LOAD SEGMENT BYTE}
 $42: Result:=1;		{1,LOAD SEGMENT REAL}
 $43: Result:=1;		{1,STORE SEGMENT INTEGER}
 $44: Result:=1;		{1,STORE SEGMENT BYTE}
 $45: Result:=1;		{1,STORE SEGMENT REAL}
 $46: Result:=1;		{1,LOAD SHORT SEGMENT REAL}
 $47: Result:=1;		{1,STORE SHORT SEGMENT REAL}
 $48: Result:=1;		{1,RESERVE MULTIDIMENSIONAL ARRAY}
 $49: Result:=1+IntBytes;	{3,Downto Type'FOR' LOOP CONTROL}
 $4A: Result:=2+IntBytes;	{4,DECREMENT AND PUSH}
 $4B: Result:=1;		{1, BAD OPCODE}
 $4C: Result:=1;		{1, BAD OPCODE}
 $4D: Result:=1;		{1, BAD OPCODE}
 $4E: Result:=1;		{1, Reserved for Port In}
 $4F: Result:=1;		{1, Reserved for Port Out}
 $50: Result:=1;		{1, Integer Absolute Value}
 $51: Result:=1;		{1, Remainder }
 $52: Result:=1;		{1, Swap Bytes }
 $53: Result:=1;		{1, Extend Sign }
 $54: Result:=1;		{1, Integer Square Root }
 $55: Result:=1;		{1, Integer Square }
 $56: Result:=1;		{1, Convert Real to Integer Number }
 $57: Result:=1;		{1, Convert Integer to Real Number }
 $58: Result:=1;		{1, Floating Point ABS }
 $59: Result:=1;		{1, Floating Point Square Root }
 $5A: Result:=1;		{1, Floating Point Square }
 $60: Result:=2;		{2, Set string termination flag}
 $61: Result:=1;		{1, Load Struct Indexed (LSX)}
 $62: Result:=2;		{2, Load Indexed Real Struct (LIRS)}
 $63: Result:=2;		{2, Load Indexed Integer Struct (LIIS) }
 $64: Result:=1;		{1, Load Indexed Byte Struct (LIBS) }
 $65: Result:=1;		{1, Store Indexed Real Struct (SIRS)}
 $66: Result:=1;		{1, Store Indexed Integer Struct (SIIS)}
 $67: Result:=1;		{1, Store Indexed Byte Struct (SIBS)}
 $68: Result:=1;		{1, Store Structure Indexed (SSX)}
 $69: Result:=1;		{1, Push Indexed Struct Address (PISA) }
 $6A: Result:=1;		{1, Push indirect struct (PINS)}
 $6B: Result:=1;		{ $6B,1, Store Structure On top of Heap (SSHP)}
 else Result:=1;		{Unknown so we'll guess 1}
 end;
end;


function GetWord(var Code: TCodeArray; Inx: integer): TI2LWord;
var I: integer;
begin
Result:=0;
for I:=0 to IntBytes-1 do
 Result:=Result or (Code[Inx+I] shl (8 * I));
end;



function IsJump(var I2l: TCodeArray; Inx: integer; var NewInx: integer; var Uncon: boolean): boolean;
{Return true if the current opcode is a jump instruction}
{If it is a jump, return the Target address in NewIndex}
{Uncon = Unconditional Jump = Jump that doesn't have two paths}
var Tmp: integer;
begin
Result:=False;
Uncon:=False;
case I2l[Inx] of
 $05:
	begin
        {Procedure call}
	Tmp:=GetWord(I2L,Inx+2);
	{An address $FFFF indicates an extermal call}
	if Tmp=$FFFF then exit;
	NewInx:=Tmp;
	end;
 $07:	begin			{Absolute Jump}
	Uncon:=True;
	NewInx:=GetWord(I2L,Inx+1);
	end;
 $08: NewInx:=GetWord(I2L,Inx+1);	{Conditional Jump}
 $18: NewInx:=GetWord(I2L,Inx+1);	{For Loop Conditional Jump}
 $25: NewInx:=GetWord(I2L,Inx+1);	{Case Jump}
 $26: NewInx:=GetWord(I2L,Inx+1);	{Optimized Procedure Call}
 $49: NewInx:=GetWord(I2L,Inx+1);	{Downto type For Loop Conditional Jump}
 else exit;
 end;
Result:=True;
end;



procedure FollowBranch(var I2l: TCodeArray; var Flags: TFlagArray; Inx: integer);
var I,Cnt,NewInx: integer;
var Op: byte;
var Stop: boolean;
begin
while true do
	begin
	{Already marked as code? then exit}
	if Flags[Inx]=false then break;
	Op:=I2L[Inx];
	Cnt:=GetOpCodeBytes(Op);
	for I:=0 to Cnt-1 do Flags[Inx+I]:=false;
	if Op=$06 then break;
	if IsJump(I2l,Inx,NewInx,Stop) then
		begin
		FollowBranch(I2l,Flags,NewInx);
		if Stop then break;
		end;
	Inc(Inx,Cnt);
	if Inx>High(I2L) then break;
	end;
end;


function PreParseCodeI2L(I2L: TCodeArray; var DataFlags: TFlagArray): string;
{Parse code looking for data areas}
var I,Inx: integer;
begin
SetLength(DataFlags,Length(I2L));
{Flag everything as data until proven otherwise}
for I:=0 to High(DataFlags) do DataFlags[I]:=true;
{First op is always an exit code}
DataFlags[0]:=false;
for I:=0 to High(SymbolDump) do
	begin
	Inx:=SymbolDump[I].BeginPC+1;
	if DataFlags[Inx] then FollowBranch(I2L,DataFlags,Inx);
	end;

end;




function DumpFlagTable(I2L: TCodeArray; Flags: TFlagArray): string;
{Convert Data/Code flags into ASCII string}
var I,J,Inx: integer;
var S: string;
const Cols = 16;
begin
Result:='';
for I:=0 to High(Flags) div Cols do
	begin
	S:=IntToHex(I*Cols,4)+': ';
	for J:=0 to Cols-1 do
		begin
		Inx:=I * Cols + J;
		if Inx>High(Flags) then break;
                S:=S+IntToHex(I2l[Inx],2)+'='+PickString(Flags[Inx],'Data, ','Code, ');
		end;
        S:=S+CRLF_Char;
	Result:=Result+S;
	end;
end;



function GetDataDump(I2L: TCodeArray; Parse: boolean): string;
{Get ASCII data dump o}
begin
if Parse then PreParseCodeI2L(I2L,FlagArray);
Result:=DumpFlagTable(I2L, FlagArray);
end;


{==============================================================================}

function DissectCode(var Code: TCodeArray; Flags: TFlagArray; Inx: integer; Args: array of TOperand): TDisasmInfo;
{Dissect an opcode and return it component parts for easy analysis}
var I,J,OldInx,Count: integer;
begin
Result.Name:='';
Result.Address:=Inx;
SetLength(Result.Operands,Length(Args));
OldInx:=Inx;
Result.Opcode:=Code[Inx];
Inc(Inx);
I:=0;
while I<=High(Args) do
	begin
	Result.Operands[I]:=Args[I];
	case Args[I] of
	 otNone: break;
	 otGlobal: Result.ByteValue:=Result.Opcode and $7F;
	 otLevel,otByteDec,otByteHex:
		begin
		Result.ByteValue:=Code[Inx];
		Inc(Inx);
		end;
	 otOffset,otAddress,otWordHex,otWordDec:
		begin
		Result.WordValue:=GetWord(Code,Inx);
		Inc(Inx,IntBytes);
		end;
	 otFloat:
		begin
		for J:=0 to RealSize-1 do Result.FloatNum[J]:=Code[Inx+J];
		Inc(Inx,RealSize);
		end;
	 otData: ;
	 else break;
	 end;
	Inc(I);
	end;
Count:=Inx-OldInx;
SetLength(Result.Data,Count);
for I:=0 to Count-1 do Result.Data[I]:=Code[OldInx+I];
end;



function DissectData(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{Dissect a data block}
var I: integer;
begin
Result.Name:='';
Result.Address:=Inx;
SetLength(Result.Operands,1);
Result.Operands[0]:=otData;
Result.Opcode:=Code[Inx];
SetLength(Result.Data,0);
I:=0;
while (Inx<Length(Flags)) and Flags[Inx] do
	begin
	SetLength(Result.Data,Length(Result.Data)+1);
	Result.Data[I]:=Code[Inx];
	Inc(I); Inc(Inx);
	end;
end;



{---------------- Convert Dissected Opcode/Data to ASCII string --------------}


function ProcessASCII(B: byte): char;
begin
if (B<$20) then Result:='.'
else Result:=char(B and $7F);
end;





function HandleDataBlock(var Info: TDisasmInfo): string;
{Extract data block and convert to hex and ASCII string}
var Cnt,I: integer;
var NewLine: boolean;
var A,S: string;
var SL: TStringList;

	procedure DumpStrings;
	begin
	SL.Add(GetTabStrLeft(S,45)+'; "'+A+'"');
	A:='';
	Inc(Info.LineCnt);
	end;


begin
SL:=TStringList.Create;
try
Result:='';
Cnt:=0;
NewLine:=True;
for I:=0 to High(Info.Data) do
	begin
	if NewLine then
		begin
                S:=IntToHex(Info.Address+I,IntNibbles)+':';
		A:='';
		Cnt:=0;
		end;
	S:=S+' '+IntToHex(Info.Data[I],2);
	A:=A+ProcessASCII(Info.Data[I]);
	Inc(Cnt);
	NewLine:=Cnt>=8;
	if NewLine then DumpStrings;
	end;
if A<>'' then DumpStrings;
Result:=SL.Text;
finally SL.Free; end;
end;



function GetFloat(Bytes: PByteArray; Inx: integer): double;
{Convert array of 8 bytes in IEEE doubel}
var PB: PByteArray;
var I: integer;
begin
PB:=PByteArray(@Result);
for I:=0 to 7 do PB[I]:=Bytes[Inx+I];
end;



procedure GetDisasmStr(var Info: TDisasmInfo);
{Convert an opcode/data item into ASCII string}
var I,J,Inx: integer;
begin
Info.DisasmStr:='';
Info.LineCnt:=0;
if Info.Operands[0]=otData then
	begin
	Info.DisasmStr:=HandleDataBlock(Info);
	exit;
	end;
Info.DisasmStr:=IntToHex(Info.Address,IntNibbles)+': '+IntToHex(Info.Opcode,2);
Inx:=1;
for I:=0 to High(Info.Operands) do
	begin
	case Info.Operands[I] of
	 otNone,otGlobal: break;
	 otByteHex,otByteDec, otLevel:
	 	begin
	 	Info.DisasmStr:=Info.DisasmStr+' '+IntToHex(Info.Data[Inx],2);
	 	Inc(Inx);
	 	end;
	 otWordHex,otWordDec,otOffset,otAddress:
		begin
		for J:=0 to IntBytes-1 do
			begin
			Info.DisasmStr:=Info.DisasmStr+' '+IntToHex(Info.Data[Inx+J],2);
			end;
		Inc(Inx,IntBytes);
		end;
	 otFloat:
		begin
		Info.DisasmStr:=Info.DisasmStr+' ';
		for J:=0 to RealSize-1 do
			begin
			Info.DisasmStr:=Info.DisasmStr+IntToHex(Info.Data[Inx+J],2);
			end;
		Inc(Inx,RealSize);
		end;
	 otData: ;
	 else break;
	 end;
	end;

if Info.Name<>'' then
	begin
	Info.DisasmStr:=GetTabStrLeft(Info.DisasmStr,30);
	Info.DisasmStr:=Info.DisasmStr+Info.Name;
	for I:=0 to High(Info.Operands) do
		begin
		case Info.Operands[I] of
		 otNone: break;
		 otByteHex,otGlobal: Info.DisasmStr:=Info.DisasmStr+' ($'+IntToHex(Info.ByteValue,2)+')';
		 otByteDec: Info.DisasmStr:=Info.DisasmStr+' ('+IntToStr(Info.ByteValue)+')';
		 otLevel: Info.DisasmStr:=Info.DisasmStr+' (Lev:'+IntToStr(Info.ByteValue)+')';
		 otWordHex: Info.DisasmStr:=Info.DisasmStr+' ($'+IntToHex(Info.WordValue and WordMask,IntNibbles)+')';
		 otAddress: Info.DisasmStr:=Info.DisasmStr+' (Add:$'+IntToHex(Info.WordValue and WordMask,IntNibbles)+')';
		 otWordDec: Info.DisasmStr:=Info.DisasmStr+' ('+IntToStr(Info.WordValue)+')';
		 otOffset: Info.DisasmStr:=Info.DisasmStr+' (Off:$'+IntToHex(Info.WordValue and WordMask,IntNibbles)+')';
		 otFloat: Info.DisasmStr:=Info.DisasmStr+'     '+FloatToStr(GetFloat(@Info.FloatNum,0));
		 else break;
		 end;
		end;
	end;
Info.DisasmStr:=Info.DisasmStr+CRLF_Char;
Info.LineCnt:=1;
end;


{------------------ Analyze individual opcodes -------------------------------}



function OPERR(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3C,1, BAD OPCODE }
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='ERROR';
Inc(Inx);
end;



function GLOBDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $8X,1, Global Load }
begin
Result:=DissectCode(Code,Flags,Inx,[otGlobal]);
Result.Name:='GLOB';
Inc(Inx);
end;



function EXITDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $00,1, EXIT INTERPRETER}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='EXIT';
Inc(Inx);
end;



function LODDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $01,4, LOAD A VARIABLE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='LOD';
Inc(Inx,2+IntBytes);
end;



function LDXDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $02,1, INDEXED LOAD A BYTE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LDX';

Inc(Inx);
end;



function STODO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $03,4, STORE INTO A VARIABLE }
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='STO';
Inc(Inx,2+IntBytes);
end;



function STXDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $04,1, INDEXED STORE TO A BYTE}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='STX';

Inc(Inx);
end;



function CALDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $05,4, CALL AN I2L PROCEDURE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otAddress]);
Result.Name:='CAL';
Inc(Inx,2+IntBytes);
end;



function RETDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $06,1, RETURN FROM I2L PROCEDURE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='RET';

Inc(Inx);
end;



function JMPDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $07,3, JUMP TO I2L CODE }
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='JMP';
Inc(Inx,1+IntBytes);
end;



function JPCDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $08,3, JUMP IF FALSE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='JPC';
Inc(Inx,1+IntBytes);
end;



function HPIDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $09,3, INCREMENT HEAP PTR BY ARG }
begin
Result:=DissectCode(Code,Flags,Inx,[otWordDec]);
Result.Name:='HPI';
Inc(Inx,1+IntBytes);
end;



function ARGDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $0A,2, GET PROC ARGUMENTS --}
begin
Result:=DissectCode(Code,Flags,Inx,[otByteDec]);
Result.Name:='ARG';
Inc(Inx,2);
end;



function IMMDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $0B,3, IMMEDIATE LOAD OF ARG }
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='IMM';
Inc(Inx,1+IntBytes);
end;



function CMLDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $0C,2, CALL AN INTRINSIC --}
begin
Result:=DissectCode(Code,Flags,Inx,[otByteDec]);
Result.Name:='CML';
Inc(Inx,2);
end;



function ADDDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $0D,1, ADD ++ --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='ADD';

Inc(Inx);
end;



function SUBDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $0E,1, SUBTRACT --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SUB';

Inc(Inx);
end;



function MULDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $0F,1, MULTIPLY --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='MUL';

Inc(Inx);
end;



function DIVDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $10,1, DIVIDE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='DIV';

Inc(Inx);
end;



function NEGDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $11,1, UNARY MINUS --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='NEG';

Inc(Inx);
end;



function EQDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $12,1, TEST FOR EQUAL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='EQ';

Inc(Inx);
end;



function NEDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $13,1, TEST FOR NOT EQUAL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='NE';

Inc(Inx);
end;



function GEDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $14,1, TEST FOR >= --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='GE';

Inc(Inx);
end;



function GTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $15,1, TEST FOR > --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='GT';

Inc(Inx);
end;



function LEDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $16,1, TEST FOR <= --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LE';

Inc(Inx);
end;



function LTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $17,1, TEST FOR < --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LT';
Inc(Inx);
end;



function FORDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $18,3, 'FOR' LOOP CONTROL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='FOR';
Inc(Inx,1+IntBytes);
end;



function INCDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $19,4, INCREMENT AND PUSH --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otAddress]);
Result.Name:='INP';
Inc(Inx,2+IntBytes);
end;



function ORDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $1A,1, OR --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='OR';

Inc(Inx);
end;



function ANDDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $1B,1, AND --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='AND';
Inc(Inx);
end;



function NOTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $1C,1, 1'S COMPLEMENT --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='NOT';
Inc(Inx);
end;



function EORDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $1D,1, EOR --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='EOR';
Inc(Inx);
end;



function DOUBL(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $1E,1, NOS+TOS*2 (DBA) --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='DBA';
Inc(Inx);
end;



function DEFSAV(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $1F,1, INDIRECT SAVE (STD) --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='STD';
Inc(Inx);
end;



function DEFER(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $20,1, INDIRECT GET (DBX) --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='DBX';
Inc(Inx);
end;



function ADDRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $21,4, ADDRESS OF VARIABLE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='ADR';
Inc(Inx,2+IntBytes);
end;



function LDIDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $22,1, INDIRECT GET --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LDI';
Inc(Inx);
end;



function LDADO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $23,3, ABSOLUTE GET --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='LDA';
Inc(Inx,2+IntBytes);
end;



function SIMMDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $24,2, SHORT IMMEDIATE }
begin
Result:=DissectCode(Code,Flags,Inx,[otByteHex]);
Result.Name:='IMS';
Inc(Inx,2);
end;



function CAJMP(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $25,3, CASE JUMP --}
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='CPJ';
Inc(Inx,1+IntBytes);
end;


function JSRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $26,3, OPTIMIZED PROC CALL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='JSR';
Inc(Inx,1+IntBytes);
end;



function RTSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $27,1, OPTIMIZED RETURN --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='RTS';
Inc(Inx);
end;



function DRPDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $28,1, DROP TOS --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='DRP';
Inc(Inx);
end;



function EXTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $29,3, CALL EXTERNAL SUBROUTINE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otAddress]);
Result.Name:='CEXT';
Inc(Inx,1+IntBytes);
end;



function FLODDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $2A,4, FLOATING LOAD --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='FLOD';
Inc(Inx,2+IntBytes);
end;



function FSTODO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $2B,4, FLOATING STORE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='FSTO';
Inc(Inx,2+IntBytes);
end;



function FIMMDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $2C,9, FLOATING LOAD IMMED --}
begin
Result:=DissectCode(Code,Flags,Inx,[otFloat]);
Result.Name:='FIMM';
Inc(Inx,1+RealSize);
end;



function FADDDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $2D,1, FLOATING ADDITION --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FADD';
Inc(Inx);
end;



function FSUBDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $2E,1, FLOATING SUBTRACT --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FSUB';

Inc(Inx);
end;



function FMULDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $2F,1, FLOATING MULTIPLY --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FMUL';

Inc(Inx);
end;



function FDIVDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $30,1, FLOATING DIVIDE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FDIV';
Inc(Inx);
end;



function FNEGDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $31,1, FLOATING NEGATE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FNEG';
Inc(Inx);
end;



function FEQDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $32,1, TEST FLOATING FOR = --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FEQ';
Inc(Inx);
end;



function FNEDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $33,1, TEST FLOATING FOR # --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FNE';
Inc(Inx);
end;



function FGEDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $34,1, TEST FLOATING FOR >= --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FGE';
Inc(Inx);
end;



function FGTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $35,1, TEST FLOATING FOR > --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FGT';
Inc(Inx);
end;



function FLEDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $36,1, TEST FLOATING FOR <= --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FLE';
Inc(Inx);
end;



function FLTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $37,1, TEST FLOATING FOR < --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FLT';
Inc(Inx);
end;



function TRADO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $38,1, TOS(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer):= 5*TOS +NOS --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='TRA';
Inc(Inx);
end;



function TRXDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $39,1, PUSH @(5*TOS+NOS) --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='TRX';
Inc(Inx);
end;



function TRIDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3A,1, PUSH(5)@(TOS(2)+NOS(5)) --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='TRI';
Inc(Inx);
end;



function STTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3B,1, PULL(5) @(NOS(2)) --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='STT';
Inc(Inx);
end;



function FADDRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3C,4, Push address of Real --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otOffset]);
Result.Name:='MARK';
Inc(Inx,2+IntBytes);
end;


function ASRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3D,1, Arithmetic ShiFt Right --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='ASR';
Inc(Inx);
end;




function LSLDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3E,1, SHIFT LEFT --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LSL';
Inc(Inx);
end;



function LSRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $3F,1, SHIFT RIGHT --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LSR';
Inc(Inx);
end;



function LDSIDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $40,1, LOAD SEGMENT INTEGER --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LDSI';
Inc(Inx);
end;



function LDSBDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $41,1, LOAD SEGMENT BYTE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LDSB';
Inc(Inx);
end;



function LDSRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $42,1, LOAD SEGMENT REAL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LDSR';
Inc(Inx);
end;



function STSIDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $43,1, STORE SEGMENT INTEGER --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='STSI';
Inc(Inx);
end;



function STSBDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $44,1, STORE SEGMENT BYTE --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='STSB';
Inc(Inx);
end;



function STSRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $45,1, STORE SEGMENT REAL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='STSR';
Inc(Inx);
end;



function LSHORT(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $46,1, LOAD SHORT SEGMENT REAL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LSHORT';
Inc(Inx);
end;



function SSHORT(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $47,1, STORE SHORT SEGMENT REAL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SSHORT';
Inc(Inx);
end;



function MKARRAY(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $48,1, RESERVE MULTIDIMENSIONAL ARRAY --}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='MKARRAY';
Inc(Inx);
end;



function FORDDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $49,3, 'Downto Type FOR' LOOP CONTROL --}
begin
Result:=DissectCode(Code,Flags,Inx,[otWordHex]);
Result.Name:='FORDOWN';
Inc(Inx,1+IntBytes);
end;



function DECDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $4A,4, INCREMENT AND PUSH --}
begin
Result:=DissectCode(Code,Flags,Inx,[otLevel,otAddress]);
Result.Name:='DEC';
Inc(Inx,2+IntBytes);
end;




function ABSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $50,1,  -- Absolutue Value}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='ABS';

Inc(Inx);
end;



function REMDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ 51$,1,  -- Remainder}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='REM';

Inc(Inx);
end;



function SWAPDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $52,1,  -- Swap}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SWAP';
Inc(Inx);
end;



function EXTENDDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $53,1,  -- Extend Sign}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='EXT';

Inc(Inx);
end;



function SQRTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $54,1,  -- Integer Square Root}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SQRT';
Inc(Inx);
end;



function SQRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $55,1,  -- Integer Square}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SQR';
Inc(Inx);
end;



function FIXDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $56,1,  -- Real to Integer (Fix)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FIX';
Inc(Inx);
end;



function FLOATDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $57,1,  -- Integer To Real (Float)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FLOAT';
Inc(Inx);
end;



function FABSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $58,1,  -- Real Absolute Value}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FABS';
Inc(Inx);
end;



function FSQRTDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $59,1,  -- Float Square Root}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FSQRT';
Inc(Inx);
end;



function FSQRDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $5A,1,  -- Float Square}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='FSQR';
Inc(Inx);
end;



function STRMODDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $60,2, Set String Termination Mode Flag --}
begin
Result:=DissectCode(Code,Flags,Inx,[otByteDec]);
Result.Name:='STRMOD';
Inc(Inx,2);
end;



function LSXDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $61,1, Load Struct Indirect }
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LSX';
Inc(Inx);
end;



function LIRSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $62,2, Load Struct Real Indexed }
begin
Result:=DissectCode(Code,Flags,Inx,[otByteDec]);
Result.Name:='LIRS';
Inc(Inx,2);
end;


function LIISDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $63,2, Load Indexed Integer Struct (LIIS) }
begin
Result:=DissectCode(Code,Flags,Inx,[otByteDec]);
Result.Name:='LIIS';
Inc(Inx,2);
end;



function LIBSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $64,1, Load Indexed Byte Struct (LIBS) }
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='LIBS';
Inc(Inx);
end;



function SIRSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $65,1, Store Indexed Real Struct (SIRS)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SIRS';
Inc(Inx);
end;


function SIISDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $66,1, Store Indexed Integer Struct (SIIS)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SIIS';
Inc(Inx);
end;



function SIBSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $67,1, Store Indexed Byte Struct (SIBS)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SIBS';
Inc(Inx);
end;


function SSXDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $68,2, Store Structure Indexed (SSX)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SSX';
Inc(Inx);
end;


function PISADO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $69,1,  Push Indexed Struct Address (PISA)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='PISA';
Inc(Inx);
end;



function PINSDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $6A,1, Push indirect struct (PINS)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='PINS';
Inc(Inx);
end;


function SSHPDO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $6B,1, Store Structure On top of Heap (SSHP)}
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='SSHP';
Inc(Inx);
end;



function PIIADO(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
{ $6C,1, Push Indexed Indirect Struct Address }
begin
Result:=DissectCode(Code,Flags,Inx,[otNone]);
Result.Name:='PIIA';
Inc(Inx);
end;



function DisassembleOpcode(var Code: TCodeArray; Flags: TFlagArray; var Inx: integer): TDisasmInfo;
var OpCode: byte;
begin
OpCode:=Code[Inx];
if Flags[Inx] then Result:=DissectData(Code,Flags,Inx)
else if (OpCode and $80)<>0 then Result:=GLOBDO(Code,Flags,Inx)
else case Opcode of
 $00: Result:=EXITDO(Code,Flags,Inx);		{ $00,1, EXIT INTERPRETER}
 $01: Result:=LODDO(Code,Flags,Inx);		{ $01,4, LOAD A VARIABLE}
 $02: Result:=LDXDO(Code,Flags,Inx);		{ $02,1, INDEXED LOAD A BYTE}
 $03: Result:=STODO(Code,Flags,Inx);		{ $03,4, STORE INTO A VARIABLE}
 $04: Result:=STXDO(Code,Flags,Inx);		{ $04,1, INDEXED STORE TO A BYTE}
 $05: Result:=CALDO(Code,Flags,Inx);		{ $05,4, CALL AN I2L PROCEDURE}
 $06: Result:=RETDO(Code,Flags,Inx);		{ $06,1, RETURN FROM I2L PROCEDURE}
 $07: Result:=JMPDO(Code,Flags,Inx);		{ $07,3, JUMP TO I2L CODE}
 $08: Result:=JPCDO(Code,Flags,Inx);		{ $08,3, JUMP IF FALSE}
 $09: Result:=HPIDO(Code,Flags,Inx);		{ $09,3, INCREMENT HEAP PTR BY ARG}
 $0A: Result:=ARGDO(Code,Flags,Inx);		{ $0A,3, GET PROC ARGUMENTS}
 $0B: Result:=IMMDO(Code,Flags,Inx);		{ $0B,3, IMMEDIATE LOAD OF ARG}
 $0C: Result:=CMLDO(Code,Flags,Inx);		{ $0C,2, CALL AN INTRINSIC}
 $0D: Result:=ADDDO(Code,Flags,Inx);		{ $0D,1, ADD}
 $0E: Result:=SUBDO(Code,Flags,Inx);		{ $0E,1, SUBTRACT}
 $0F: Result:=MULDO(Code,Flags,Inx);		{ $0F,1, MULTIPLY}
 $10: Result:=DIVDO(Code,Flags,Inx);		{ $10,1, DIVIDE}
 $11: Result:=NEGDO(Code,Flags,Inx);		{ $11,1, UNARY MINUS}
 $12: Result:=EQDO(Code,Flags,Inx);		{ $12,1, TEST FOR EQUAL}
 $13: Result:=NEDO(Code,Flags,Inx);		{ $13,1, TEST FOR NOT EQUAL}
 $14: Result:=GEDO(Code,Flags,Inx);		{ $14,1, TEST FOR >=}
 $15: Result:=GTDO(Code,Flags,Inx);		{ $15,1, TEST FOR >}
 $16: Result:=LEDO(Code,Flags,Inx);		{ $16,1, TEST FOR <=}
 $17: Result:=LTDO(Code,Flags,Inx);		{ $17,1, TEST FOR <}
 $18: Result:=FORDO(Code,Flags,Inx);		{ $18,3, 'FOR' LOOP CONTROL}
 $19: Result:=INCDO(Code,Flags,Inx);		{ $19,4, INCREMENT AND PUSH}
 $1A: Result:=ORDO(Code,Flags,Inx);		{ $1A,1, OR}
 $1B: Result:=ANDDO(Code,Flags,Inx);		{ $1B,1, AND}
 $1C: Result:=NOTDO(Code,Flags,Inx);		{ $1C,1, 1'S COMPLEMENT}
 $1D: Result:=EORDO(Code,Flags,Inx);		{ $1D,1, EOR}
 $1E: Result:=DOUBL(Code,Flags,Inx);		{ $1E,1, NOS+TOS*2 (DBA)}
 $1F: Result:=DEFSAV(Code,Flags,Inx);		{ $1F,1, INDIRECT SAVE (STD)}
 $20: Result:=DEFER(Code,Flags,Inx);		{ $20,1, INDIRECT GET (DBX)}
 $21: Result:=ADDRDO(Code,Flags,Inx);		{ $21,4, ADDRESS OF VARIABLE}
 $22: Result:=LDIDO(Code,Flags,Inx);		{ $22,1, INDIRECT GET}
 $23: Result:=LDADO(Code,Flags,Inx);		{ $23,3, ABSOLUTE GET}
 $24: Result:=SIMMDO(Code,Flags,Inx);		{ $24,2, SHORT IMMEDIATE}
 $25: Result:=CAJMP(Code,Flags,Inx);		{ $25,3, CASE JUMP}
 $26: Result:=JSRDO(Code,Flags,Inx);		{ $26,3, OPTIMIZED PROC CALL}
 $27: Result:=RTSDO(Code,Flags,Inx);		{ $27,1, OPTIMIZED RETURN}
 $28: Result:=DRPDO(Code,Flags,Inx);		{ $28,1, DROP TOS}
 $29: Result:=EXTDO(Code,Flags,Inx);		{ $29,3, CALL EXTERNAL SUBROUTINE}
 $2A: Result:=FLODDO(Code,Flags,Inx);		{ $2A,4, FLOATING LOAD}
 $2B: Result:=FSTODO(Code,Flags,Inx);		{ $2B,4, FLOATING STORE}
 $2C: Result:=FIMMDO(Code,Flags,Inx);		{ $2C,9, FLOATING LOAD IMMED}
 $2D: Result:=FADDDO(Code,Flags,Inx);		{ $2D,1, FLOATING ADDITION}
 $2E: Result:=FSUBDO(Code,Flags,Inx);		{ $2E,1, FLOATING SUBTRACT}
 $2F: Result:=FMULDO(Code,Flags,Inx);		{ $2F,1, FLOATING MULTIPLY}
 $30: Result:=FDIVDO(Code,Flags,Inx);		{ $30,1, FLOATING DIVIDE}
 $31: Result:=FNEGDO(Code,Flags,Inx);		{ $31,1, FLOATING NEGATE}
 $32: Result:=FEQDO(Code,Flags,Inx);		{ $32,1, TEST FLOATING FOR =}
 $33: Result:=FNEDO(Code,Flags,Inx);		{ $33,1, TEST FLOATING FOR #}
 $34: Result:=FGEDO(Code,Flags,Inx);		{ $34,1, TEST FLOATING FOR >=}
 $35: Result:=FGTDO(Code,Flags,Inx);		{ $35,1, TEST FLOATING FOR >}
 $36: Result:=FLEDO(Code,Flags,Inx);		{ $36,1, TEST FLOATING FOR <=}
 $37: Result:=FLTDO(Code,Flags,Inx);		{ $37,1, TEST FLOATING FOR <}
 $38: Result:=TRADO(Code,Flags,Inx);		{ $38,1, TOS:= 5*TOS +NOS}
 $39: Result:=TRXDO(Code,Flags,Inx);		{ $39,1, PUSH @(5*TOS+NOS)}
 $3A: Result:=TRIDO(Code,Flags,Inx);		{ $3A,1, PUSH(5)@(TOS(2)+NOS(5))}
 $3B: Result:=STTDO(Code,Flags,Inx);		{ $3B,1, PULL(5) @(NOS(2))}
 $3C: Result:=FADDRDO(Code,Flags,Inx);		{ $3C,1, Push real address}
 $3D: Result:=ASRDO(Code,Flags,Inx);		{ $3D,1, Arithmetic Shift Right}
 $3E: Result:=LSLDO(Code,Flags,Inx);		{ $3E,1, SHIFT LEFT}
 $3F: Result:=LSRDO(Code,Flags,Inx);		{ $3F,1, SHIFT RIGHT}
 $40: Result:=LDSIDO(Code,Flags,Inx);		{ $40,1, LOAD SEGMENT INTEGER}
 $41: Result:=LDSBDO(Code,Flags,Inx);		{ $41,1, LOAD SEGMENT BYTE}
 $42: Result:=LDSRDO(Code,Flags,Inx);		{ $42,1, LOAD SEGMENT REAL}
 $43: Result:=STSIDO(Code,Flags,Inx);		{ $43,1, STORE SEGMENT INTEGER}
 $44: Result:=STSBDO(Code,Flags,Inx);		{ $44,1, STORE SEGMENT BYTE}
 $45: Result:=STSRDO(Code,Flags,Inx);		{ $45,1, STORE SEGMENT REAL}
 $46: Result:=LSHORT(Code,Flags,Inx);		{ $46,1, LOAD SHORT SEGMENT REAL}
 $47: Result:=SSHORT(Code,Flags,Inx);		{ $47,1, STORE SHORT SEGMENT REAL}
 $48: Result:=MKARRAY(Code,Flags,Inx);		{ $48,1, RESERVE MULTIDIMENSIONAL ARRAY}
 $49: Result:=FORDDO(Code,Flags,Inx);		{ $49,3, 'Downto Type FOR' LOOP CONTROL}
 $4A: Result:=DECDO(Code,Flags,Inx);		{ $4A,4, Decrement AND PUSH}
 $4B: Result:=OPERR(Code,Flags,Inx);		{ $4B,1, BAD OPCODE}
 $4C: Result:=OPERR(Code,Flags,Inx);		{ $4C,1, BAD OPCODE}
 $4D: Result:=OPERR(Code,Flags,Inx);		{ $4D,1, BAD OPCODE}
 $4E: Result:=OPERR(Code,Flags,Inx);		{ $4E,1, Reserved for Port In}
 $4F: Result:=OPERR(Code,Flags,Inx);		{ $4F,1, Reserved for Port Out}
 $50: Result:=ABSDO(Code,Flags,Inx);		{ $50,1, Integer Absolute Value}
 $51: Result:=REMDO(Code,Flags,Inx);		{ $51,1, Remainder }
 $52: Result:=SWAPDO(Code,Flags,Inx);		{ $52,1, Swap Bytes }
 $53: Result:=EXTENDDO(Code,Flags,Inx);		{ $53,1, Extend Sign }
 $54: Result:=SQRTDO(Code,Flags,Inx);		{ $54,1, Integer Square Root }
 $55: Result:=SQRDO(Code,Flags,Inx);		{ $55,1, Integer Square }
 $56: Result:=FIXDO(Code,Flags,Inx);		{ $56,1, Convert Real to Integer Number }
 $57: Result:=FLOATDO(Code,Flags,Inx);		{ $57,1, Convert Integer to Real Number }
 $58: Result:=FABSDO(Code,Flags,Inx);		{ $58,1, Floating Point ABS }
 $59: Result:=FSQRTDO(Code,Flags,Inx);		{ $59,1, Floating Point Square Root }
 $5A: Result:=FSQRDO(Code,Flags,Inx);		{ $5A,1, Floating Point Square }
 $60: Result:=STRMODDO(Code,Flags,Inx);		{ $60,2, Set String Termination Mode Flag }
 $61: Result:=LSXDO(Code,Flags,Inx);		{ $61,1, Load Struct Indirect (LSX)}
 $62: Result:=LIRSDO(Code,Flags,Inx);		{ $62,2, Load Indexed Real Struct (LIRS)}
 $63: Result:=LIISDO(Code,Flags,Inx);		{ $63,2, Load Indexed Integer Struct (LIIS) }
 $64: Result:=LIBSDO(Code,Flags,Inx);		{ $64,1, Load Indexed Byte Struct (LIBS) }
 $65: Result:=SIRSDO(Code,Flags,Inx);		{ $65,1, Store Indexed Real Struct (SIRS)}
 $66: Result:=SIISDO(Code,Flags,Inx);		{ $66,1, Store Indexed Integer Struct (SIIS)}
 $67: Result:=SIBSDO(Code,Flags,Inx);		{ $67,1, Store Indexed Byte Struct (SIBS)}
 $68: Result:=SSXDO(Code,Flags,Inx);		{ $68,1, Store Structure Indexed (SSX)}
 $69: Result:=PISADO(Code,Flags,Inx);           { $69,1, Push Indexed Struct Address (PISA) }
 $6A: Result:=PINSDO(Code,Flags,Inx);		{ $6A,1, Push indirect struct (PINS)}
 $6B: Result:=SSHPDO(Code,Flags,Inx);		{ $6B,1, Store Structure On top of Heap (SSHP)}
 $6C: Result:=PIIADO(Code,Flags,Inx);		{ $6C,1, Push Indexed Indirect Struct Address }
 else Result:=OPERR(Code,Flags,Inx);
 end;
GetDisasmStr(Result);
end;


var DisasmList: TDisasmList;


function PCToDisasmLine(PC: integer): integer;
{Find the Disassembly Line Number that corresponds to a specific PC}
var Add: integer;
var L,H,I: integer;
begin
L:=0; H:=DisasmList.Count-1;
while L <= H do
	begin
	I:=(L + H) shr 1;
	Add:=DisasmList[I].Address;
	if PC<Add then H:=I-1
	else if PC>Add then L:=I+1
	else break;
	end;
Result:=DisasmList[I].LineNum;
end;




procedure DisAssembleI2L(I2L: TCodeArray);
{Disasemble I2L Code. Result is in Global "DisasmList"}
var Info: TDisasmInfo;
var LineNum: integer;
var ProgCntr: integer;
begin
PreParseCodeI2L(I2L,FlagArray);
ProgCntr:=0;
LineNum:=0;
DisasmList.Clear;
while ProgCntr<Length(I2L) do
	begin
	Info:=DisassembleOpcode(I2L,FlagArray,ProgCntr);
	Info.LineNum:=LineNum;
	DisasmList.Add(Info);
	LineNum:=LineNum + Info.LineCnt;
	end;
end;




procedure DisAssembleI2L(I2L: TCodeArray; var Disasm: string);
{Diassemble code and return a string of the disassembly}
var I: integer;
begin
DisAsm:='';
DisAssembleI2L(I2L);
for I:=0 to DisasmList.Count-1 do
	begin
	Disasm:=Disasm+DisasmList[I].DisasmStr;
	end;
end;



procedure DisAsmI2LWithSource(I2LBin: TI2LProgram; Source: string; var Disasm: string);
{Disassemble the code and return a string of the disassembly the includes the source}
var I,J,Line,OldLine: integer;
var PCInfo: TPCInfo;
var SL: TStringList;
begin
SL:=TStringList.Create;
try
SL.Text:=Source;
DisAsm:='';
DisAssembleI2L(I2LBin.Code);
OldLine:=-1;
for I:=0 to DisasmList.Count-1 do
	begin
	PCInfo:=PCToSourceData(DisasmList[I].Address,I2LBin.BaseAddress);
	Line:=PCInfo.Info.Line;
	if Line<>OldLine then
		begin
                Disasm:=Disasm+CRLF_Char;
		for J:=OldLine+1 to Line do Disasm:=Disasm+SL[J]+CRLF_Char;
                Disasm:=Disasm+CRLF_Char;
		end;
	Disasm:=Disasm+DisasmList[I].DisasmStr;
	OldLine:=Line;
	end;
finally SL.Free; end;
end;





initialization
DisasmList:=TDisasmList.Create;
finalization
DisasmList.Free;
end.
