unit I2LGlobal;

interface

uses Windows,Types,Forms,SysUtils,Classes,DSupport,StreamLib, XPLDataTypes, I2lDataTypes,ConsoleWindow;


type TI2LError = (ieIntDivision=1, ieMemory=2, ieIOError=3, ieOpcode=4,
		  ieIntrinsic=5, ieRealDivision=6, ieOverflow=7, ieUnderflow=8,
		  ieFixError=9, ieSquareroot=10, ieLogError=11, ieExpError=12,
		  ieStackHeap=13, ieArcTan=14, ieUnknown=15);


var GlobalConsole: TConsoleWindow;		{Global console window, must be initialized}


var Remainder: TI2LWord;	{Remained from last integer divide}
var ReRunFlag: boolean;		{Flags program has been restarted}
var CloseOnTerminate: boolean;	{Flags that a stand-alone program window should close when it terminates }
var TrapControlC: boolean;	{Flags Ctrl-C will not abort program - currently unused}
var ControlCHit: boolean;	{Flags Ctrl-C was hit - currently unimplimented}
var RunFromDOS: boolean;	{Flags that the program has been run from a command prompt}


{ Program Storage array}

var ProgramStore: TI2LDataBlock;

type TStackInfo = record
 Item: TI2LWord;
 SubCall: boolean;
 end;

const MaxStackSize = 1024 * 1024 * 100;
var I2LStack: array of TStackInfo;
var StackPtr: integer;

var StackSize: integer;

{Variable to handle stack dump}

type TStackItem = record
 Address,Column,Line: integer;
 Filename,ProcName: string;
 end;

type TStackDump = array of TStackItem;

var StackDump: TStackDump;


{ For Error Hanling }

var Traps: TI2LWord;
var ErrorNum: TI2LError;

{Disk I/O Varibles}

var INHAND: integer = -1;	{Current Input file handle}
var OTHAND: integer = -1;	{Current output file handle}


{DOS-style Redirection handler varialbles}

type TRedirCommand = (rdOutput,rdAppend,rdInput,rdPipe,rdNone);

type TRedirInfo = record
 Cmd: TRedirCommand;
 Target: string;
 end;

type TRedirArray = array of TRedirInfo;

var RedirInput: TStream;
var RedirOutput: TStream;


procedure GlobalInitialize(Console: TConsoleWindow);


function FileRewind(Handle: integer): boolean;

procedure PushStack(V: TI2LWord);
procedure PushSubCall(V: TI2LWord);
function PopStack: TI2LWord;
function GetStackOffset(OffSet: integer): TI2LWord;
function PopRealPointer: TI2LWord;
procedure PushDouble(D: double);
function PopDouble: double;
procedure PushSingle(D: single);
function PopSingle: single;
function GetXPLBoolean(B: boolean): TI2LWord;
procedure PushXPLBoolean(B: boolean);
procedure GetStackDump(var StackDump: TStackDump);

procedure HandleI2LError(Descript: string; Num: TI2LError);
function HandleFloatErrors: boolean;

procedure ReserveMemory(Size: integer);

procedure InitalizeRedirection;
procedure FinalizeRedirection;
procedure ResetRedirection;
function GetRedirectInput(var C: char): boolean;
function RedirectOutput(C: char): boolean;
function RedirectString(s: string): boolean;


implementation

uses I2LInterp, I2lIntrinsics, IDEDebugInfo;



{---------- File I/O Routines -----------------------------------}

function FileRewind(Handle: integer): boolean;
begin
Result:=Handle>=0;
if Result then Result:=FileSeek(Handle,0,0)>=0;
end;





{-------------- Handle Errors -----------------------------------------}

const ErrorMask: array [0..15] of Word =(1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768);

{

Error trapping works slighlty differently from the Original I2L.
Instead of error number being tied to an individual error message,
more specific error message strings are specified at the point of
the error. The error numbers are only used for Trapping.  Thus, I2L
Errors are divided into 15 general categories. Each of these
categories can be trapped or ignored. The errors numbers are the
same and the traps bits haven't changeed, so your code doesn't have
to change. It is just a change in paradigm that gives more specific
error messages. For example, there several different situations that
can cause a memory error including Ordinary Reserves and Real reserves.
The error message now distinguish between all the different variations.

The general categories are:

1 = Integer Division
2 = Memory Errors
3 = I/O Errors
4 = I2L Opcode Errors
5 = Intrinsic Errors
6 = Real Division
7 = Real Underflow Errors
8 = Real Overflow Errors
9 = Fix Errors
10  = Square Root errors
11 = Log/LN Errors
12 = Exp Errors
13 = Stack/Heap Errors
14 = Arc Tangent Errors
15 = Uknown, unused.

}



procedure GetStackDump(var StackDump: TStackDump);
var I: integer;
begin
if I2LProgram=nil then exit;
SetLength(StackDump,1);
StackDump[High(StackDump)].Address:=I2LProgram.ProgCntr;

for I:=StackPtr-1 downto 0 do
 if I2LStack[I].SubCall then
	begin
	SetLength(StackDump,Length(StackDump)+1);
	StackDump[High(StackDump)].Address:=I2LStack[I].Item;
	end;
end;




procedure HandleI2LError(Descript: string; Num: TI2LError);
{I2L Error Handler}
var TT: TI2LWord;
begin
ErrorNum:=Num;
TT:=Traps;
if (Traps and ErrorMask[Integer(Num)])<>0 then
	begin
	raise Exception.Create(Descript);
	end;
end;




function HandleFloatErrors: boolean;
var Status: word;
begin
Result:=False;
Status:=Get8087Status(True) and $1C;
if Status = 0 then exit;
if (Status and $04)<>0 then HandleI2LError('Real Div by Zero',ieRealDivision)
else if (Status and $08)<>0 then HandleI2LError('Real Overflow',ieOverflow)
else HandleI2LError('Real Underflow',ieUnderflow);
Result:=True;
end;


{-------------- Stack simulation code -----------------------------------------}



procedure SetStackSize(Size: integer);
begin
StackSize:=Size;
SetLength(I2LStack,StackSize);
end;



procedure GrowStack;
{expand stack if necessary}
var Len: integer;
begin
SetStackSize(StackSize+StackSize div 2);
if StackPtr>=Length(I2LStack) then HandleI2LError('Stack Overflow',ieStackHeap);
end;



procedure PushStack(V: TI2LWord);
{Push word on simulated stack}
begin
if StackPtr>=StackSize then GrowStack;
I2LStack[StackPtr].Item:=V;
I2LStack[StackPtr].SubCall:=False;
Inc(StackPtr);
end;



procedure PushSubCall(V: TI2LWord);
{Push return address for subroutine call}
begin
if StackPtr>=StackSize then GrowStack;
I2LStack[StackPtr].Item:=V;
I2LStack[StackPtr].SubCall:=True;
Inc(StackPtr);
end;


function PopStack: TI2LWord;
{Pop word from simulated stack}
begin
if StackPtr<=0 then HandleI2LError('Stack Underflow',ieStackHeap);
Dec(StackPtr);
Result:=I2LStack[StackPtr].Item;
end;


function GetStackOffset(OffSet: integer): TI2LWord;
{Get item offset from top of stack}
var Inx: integer;
begin
Inx:=(StackPtr-1)-Offset;
if Inx<=0 then HandleI2LError('Stack Underflow',ieStackHeap);
Result:=I2LStack[Inx].Item;
end;


function PopRealPointer: TI2LWord;
{Pop integer pointer part of real pointer}
begin
PopStack;
Result:=PopStack;
end;



procedure PushDouble(D: double);
var I: integer;
var A: PI2LWordArray;
begin
A:=PI2LWordArray(@D);
for I:=0 to IntsPerReal-1 do PushStack(A[I]);
end;



function PopDouble: double;
var I: integer;
var A: PI2LWordArray;
begin
A:=PI2LWordArray(@Result);
for I:=IntsPerReal-1 downto 0 do A[I]:=PopStack;
end;



procedure PushSingle(D: single);
var I: integer;
var A: PI2LWordArray;
const Steps = SingleSize div IntBytes;
begin
A:=PI2LWordArray(@D);
for I:=0 to Steps-1 do PushStack(A[I]);
end;



function PopSingle: single;
var I: integer;
var A: PI2LWordArray;
const Steps = SingleSize div IntBytes;
begin
A:=PI2LWordArray(@Result);
for I:=Steps-1 downto 0 do A[I]:=PopStack;
end;


function GetXPLBoolean(B: boolean): TI2LWord;
begin
Result:=PickInteger(B,XPLTrue,XPLFalse);
end;



procedure PushXPLBoolean(B: boolean);
begin
PushStack(GetXPLBoolean(B));
end;

procedure GlobalInitialize(Console: TConsoleWindow);
begin
ReRunFlag:=false;
CloseOnTerminate:=False;
GlobalConsole:=Console;
SetStackSize(8 * 1024);
end;



procedure ReserveMemory(Size: integer);
{Handle memory reserve and throw error if it fails}
begin
if not I2LProgram.ReserveMemory(Size) then
  HandleI2LError('Reserve Exceeds Memory',ieMemory);
end;


{--------------------- DOS-style redirection handler --------------------------}


{
> = Redirect output
>> = Append ouput
< = Redirect input
| = Pipe
}



function RedirCommandToStr(Cmd: TRedirCommand): string;
begin
case Cmd of
 rdOutput: Result:='Output';
 rdAppend: Result:='Append';
 rdInput: Result:='Input';
 rdPipe: Result:='Pipe';
 else Result:='Unknown';
 end;
end;


function ParseCommandline(ComArgs: TStringList): TRedirArray;
{Parse command line parameter return redirection information}
var S,Target: string;
var I: integer;

	procedure InsertCmd(Cmd: TRedirCommand; Target: string);
	begin
	SetLength(Result,Length(Result)+1);
	Result[High(Result)].Cmd:=Cmd;
	Result[High(Result)].Target:=Target;
	end;


	procedure HandleArguments(Cmd: TRedirCommand; ArgCnt: integer);
	var Target: string;
	begin
	if Length(S)>ArgCnt then Target:=Copy(S,ArgCnt+1,Length(S)-ArgCnt)
	else if I<=ComArgs.Count-1 then
		begin
		Target:=ComArgs[I];
		Inc(I);
		end;
	InsertCmd(Cmd,Target);
	end;



begin
I:=1;
while I<=ComArgs.Count-1 do
	begin
	S:=ComArgs[I];
	Inc(I);
	if (S[1]='>') and (S[2]='>') then HandleArguments(rdAppend,2)
	else if S[1]='>' then HandleArguments(rdOutput,1)
	else if S[1]='<' then HandleArguments(rdInput,1)
	else if S[1]='|' then HandleArguments(rdPipe,1)
	end;
end;



procedure HandleWindowsMode;
{Handle running under windows and not DOS}
var RA: TRedirArray;
var I: integer;
var LastInput,LastOutput: TRedirInfo;
var S: string;
begin
RA:=ParseCommandline(ComArgs);
LastInput.Cmd:=rdNone; LastOutput.Cmd:=rdNone;
for I:=High(RA) downto 0 do
 if RA[I].Cmd in [rdOutput,rdAppend] then
	begin
	LastOutput:=RA[I];
	break;
	end;

for I:=High(RA) downto 0 do
 if RA[I].Cmd=rdInput then
	begin
	LastInput:=RA[I];
	break;
	end;
if LastInput.Cmd<>rdNone then
	begin
	RedirInput:=TMemoryStream.Create;
	TMemoryStream(RedirInput).LoadFromFile(LastInput.Target);
	RedirInput.Seek(0,soFromBeginning);
	end;
if LastOutput.Cmd<>rdNone then
	begin
	case LastOutput.Cmd of
	 rdAppend:
		begin
		if FileExists(LastOutput.Target) then RedirOutput:=TFileStream.Create(LastOutput.Target,fmOpenReadWrite)
		else RedirOutput:=TFileStream.Create(LastOutput.Target,fmCreate);
		RedirOutput.Seek(0,soEnd);
		end;
	 rdOutput:
		begin
		RedirOutput:=TFileStream.Create(LastOutput.Target,fmCreate);
		RedirOutput.Seek(0,soBeginning);
		end;
	 end;
	end;
end;



procedure HandleDOSMode;
begin
if IsOutputRedirected then
	begin
	RedirOutput:=THandleStream.Create(GetStdHandle(STD_OUTPUT_HANDLE));
	end;

if IsInputRedirected then
	begin
	RedirInput:=THandleStream.Create(GetStdHandle(STD_INPUT_HANDLE));
	RedirInput.Seek(0,soBeginning);
	end;
end;



procedure InitalizeRedirection;
begin
if RunFromDOS then HandleDOSMode
else HandleWindowsMode;
end;



procedure ResetRedirection;
begin
if (RedirInput<>nil) and (RedirInput.Size>0) then
 RedirInput.Seek(0,soBeginning);
end;



function GetRedirectInput(var C: char): boolean;
begin
Result:=(RedirInput<>nil) and (RedirInput.Size>0);
if Result then
	begin
	if RedirInput.Position>=RedirInput.Size then C:=CtrlZ
	else RedirInput.Read(C,1);
	end;
end;



function RedirectOutput(C: char): boolean;
begin
Result:=RedirOutput<>nil;
if Result then RedirOutput.Write(C,1);
end;



function RedirectString(S: string): boolean;
begin
Result:=RedirOutput<>nil;
if Result then WriteStreamString(RedirOutput,S);
end;




procedure FinalizeRedirection;
begin
FreeAndNil(RedirInput);
FreeAndNil(RedirOutput);
end;




initialization
ProgramStore:=TI2LDataBlock.Create;
finalization
RedirInput.Free;
if RedirOutput<>nil then RedirOutput.Free;
ProgramStore.Free;
end.
