unit I2LInterp;

interface

uses Windows,Messages,Forms,SysUtils,Graphics,DSupport,Math,I2LDataTypes,
  XPLDataTypes,ConsoleWindow;


var I2LProgram: TI2LDataBlock;

var LastPC: integer;		{Program Counter of last instruction}
var OpCode: byte;		{Current Opcode}
var StopProgram: boolean;	{Used to abort program}
var PauseFlag: boolean;
var UseMSBStrings: boolean;	{Flags MSB terminated string vs null Terminated}
var LoopCount: integer;		{Used by the interpreter Loop to periodically process messages}

var MemManager: TMemoryManager;


{-------------- Display Vector Simulation code --------------------------------}

const StaticNest = 8;

var DisplayVector: array [0..StaticNest-1] of TI2LWord;
var DisplayPtr: integer;


procedure RunProgram(Source: string; Console: TConsoleWindow);
procedure InitializeInterpreter(Console: TConsoleWindow);
procedure StartProgram(Console: TConsoleWindow);
procedure FinalizeInterpreter;
procedure KillProgram;
function InterpretLoop(OneStep: boolean): boolean;
procedure SetProgramCode(var Code: TI2LDataBlock);


implementation

uses I2lIntrinsics, I2LLoader, I2LGlobal, I2LConsoleIO;



{-----------------------------------------------------------------------------}




procedure OPERR;
{ BAD OPCODE}
{OPCODE ERROR DETECTED}
begin
HandleI2LError('Illegal Opcode',ieOpcode);
end;


procedure EXITDO;
{ $00,1, Exit Interpreter - not actually used}
begin
StopProgram:=True;
end;



procedure LODDO;
{ $01,4 - Routine to retrive a variable's value and push it on the stacK}
{1 =  The opcode}
{2 = The level in the display vector of the base address}
{3..WordSize = The offset from that base address of the actual value}
var Level,Offset: TI2LWord;
begin
Level:=DisplayVector[I2LProgram.GetProgramByte];
Offset:=I2LProgram.GetProgramWord;
PushStack(I2LProgram.GetDataWord(Level + Offset));
end;


procedure LDXDO;
{ $02,1 - This routine forms tos+nos and then uses that value as an}
{indirect address of a byte to push onto tos (ldx). One-byte}
{instruction.}
var W: TI2LWord;
begin
W:=I2LProgram.DataByte[PopStack+PopStack];
PushStack(W);
end;


procedure STODO;
{ $03,4 - Routine to store top of stack (tos) into a variable}
{1 =  The opcode}
{2 = The level in the display vector of the base address}
{3 = The offset from that base address of the actual value}
var Base,Offset: TI2LWord;
begin
Base:=DisplayVector[I2LProgram.GetProgramByte];
Offset:=I2LProgram.GetProgramWord;
I2LProgram.SetDataWord(Base + Offset,PopStack);
end;



procedure STXDO;
{ $04,1, INDEXED STORE TO A BYTE}
{ Store low byte of tos into address in nos and pop both (STX). One-byte}
{instruction.}
var V: TI2LWord;
begin
V:=PopStack;
I2LProgram.DataByte[PopStack]:=V;
end;





procedure CALDO;
{ $05,4, CALL AN I2L PROCEDURE}
{ After a procedure call the stack is set up as follows:}
{	Old level}
{	Old display vector at new level-low byte}
{	Old display vector at new level-high byte}
{	Return pc - low byte}
{	Return pc - high byte}

{The display vector for the new level is set to the current}
{heap location.}
{The instruction consists of:}
{1 =	the opcode}
{2 =	the level of the procedure to be invoked}
{3 =	 entry address}
var Level: TI2LWord;
begin
Level:=I2LProgram.GetProgramByte;
PushStack(Level);
PushStack(DisplayVector[Level]);
PushSubCall(I2LProgram.ProgCntr+IntBytes);
DisplayVector[Level]:=I2LProgram.HeapPtr;
I2LProgram.ProgCntr:=I2LProgram.GetProgramWord;
end;



procedure RETDO;
{ $06,1, RETURN FROM I2L PROCEDURE}
{ Routine to return from a procedure. This pops the stuff}
{pushed by the call. It restores the heap to the depth it had}
{at the time of the call. This is a one-byte instruction.}
var Level,OldLevel: TI2LWord;
begin
I2LProgram.ProgCntr:=PopStack;
Level:=PopStack;
OldLevel:=PopStack;
I2LProgram.HeapPtr:=DisplayVector[OldLevel];
DisplayVector[OldLevel]:=Level;
end;


procedure JMPDO;
{ $07,3, JUMP TO I2L CODE}
{the two jumps are each three-byte instructions.}
{1 =	Opcode}
{2.. WordSize =	target address}
begin
I2LProgram.ProgCntr:=I2LProgram.GetProgramWord;
end;


procedure JPCDO;
{ $08,3, JUMP IF FALSE}
{Note false is zero; non-zero is true. This routine jumps on}
{false, not on true. The result of the last boolean expression}
{is on the stack. It is checked and the appropriate is action taken.}
begin
if PopStack=0 then I2LProgram.ProgCntr:=I2LProgram.GetProgramWord
else Inc(I2LProgram.ProgCntr,IntBytes);
end;


procedure HPIDO;
{ $09,3, INCREMENT HEAP PTR BY ARG}
{This routine is used to reserve heap space for local}
{variables and arguments at run time. Two-byte instruction:}
{1 = opcode}
{2..WordSize = number of bytes to skip over}
begin
ReserveMemory(I2LProgram.GetProgramWord);
end;


procedure ARGDO;
{ $0A,3, GET PROC ARGUMENTS}
{routine to retrieve procedure arguments from the stack}
{and put them on the heap. The called procedure will then}
{reserve the space with an "hpi" and the variables thus}
{created will be preset to their appropriate values.}
{Two-byte instruction:}
{1 = Opcode}
{2..WordSize = Number of bytes of arguments minus one.}
var Count, I: integer;
begin
Count:=(I2LProgram.GetProgramByte+1) div IntBytes;
I2LProgram.ExpandMemory(I2LProgram.HeapPtr+Count*IntBytes);
for I:=Count-1 downto 0 do
	begin
	I2LProgram.SetDataWord(I2LProgram.HeapPtr+I*IntBytes,PopStack);
	end;
end;


procedure IMMDO;
{ $0B,3, IMMEDIATE LOAD OF ARG}
{ immediate load of a Word-Size constant onto the stack}
{three bytes:}
{1 = Opcode}
{2..WordSize = constant}
begin
PushStack(I2LProgram.GetProgramWord);
end;



procedure CMLDO;
{ $0C,2, CALL AN INTRINSIC}
{Routine to call an intrinsic. Note that intrinsics are not}
{subroutines--they return by a direct jump to "CMLRET".}
{Arguments, if any, are on the stack in the order they are}
{called (TOS is last). If an intrinsic returns a value, it will}
{be in tos. Two-byte instruction:}
{1 = opcode}
{2 = intrinsic number}
begin
CallIntrinsic(I2LProgram.GetProgramByte);
{Make sure there are no floating point errors after context switch}
Clear8087Status;
end;




procedure ADDDO;
{ $0D,1, ADD}
{TOS:=NOS+TOS}
begin
PushStack(PopStack+PopStack);
end;


procedure SUBDO;
{ $0E,1, SUBTRACT}
{TOS:=NOS-TOS}
begin
PushStack(-PopStack+PopStack);
end;


procedure MULDO;
{ $0F,1, MULTIPLY}
{TOS:=NOS*TOS}
begin
PushStack(PopStack*PopStack);
end;



procedure DIVDO;
{ $10,1, DIVIDE}
{TOS:=NOS/TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
if Tos=0 then
	begin
	HandleI2LError('Integer Div by Zero',ieIntDivision);
	Remainder:=0;
	PushStack(High(TI2LWord));
	end
else
	begin
	Remainder:=Nos mod Tos;
	PushStack(Nos div Tos);
	end;
end;



procedure NEGDO;
{ $11,1, UNARY MINUS}
{TOS:= -TOS}
begin
PushStack(-PopStack);
end;



procedure EQDO;
{ $12,1, TEST FOR EQUAL}
{TOS:= TOS = NOS}
begin
if PopStack = PopStack then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;



procedure NEDO;
{ $13,1, TEST FOR NOT EQUAL}
{TOS:= TOS # NOS}
begin
if PopStack <> PopStack then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;


procedure GEDO;
{ $14,1, TEST FOR >=}
{TOS:= NOS >= TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
if Nos >= Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;


procedure GTDO;
{ $15,1, TEST FOR >}
{TOS:= NOS > TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
if Nos > Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;


procedure LEDO;
{ $16,1, TEST FOR <=}
{TOS:= NOS <= TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
if Nos <= Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;


procedure LTDO;
{ $17,1, TEST FOR <}
{TOS:= NOS < TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
if Nos < Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;

{	*** FOR LOOP CONTROL ***}

procedure FORDO;
{ $18,3, 'FOR' LOOP CONTROL}
{This routine handles the test and branch of the 'FOR' loop.}
{The stack is assumed to contain the limit (nos) and a copy of}
{the loop control variable (tos). They are compared and if the}
{loop variable is greater than the limit, this opcode's address}
{is branched to, and the stack is cleaned up (the 'FOR' loop is}
{done). Otherwise, the PC is advanced to the next opcode, and}
{the limit value is left on the stack (the 'FOR' loop}
{continues). The code for this instruction is 3 Bytes long:}
{1 = The Opcode}
{2..WordSize = Branch Address}
begin
if PopStack>I2LStack[StackPtr-1].Item then
	begin
	{Skip past for-block and clean stack}
	I2LProgram.ProgCntr:=I2LProgram.GetProgramWord;
	PopStack;
	end
else
	begin
	{Skip the arguments}
	Inc(I2LProgram.ProgCntr,IntBytes);
	end;
end;



procedure INCDO;
{ $19,4, INCREMENT AND PUSH}
{ Instruction to Increment a variable's value.}
{It leaves a copy of the variable on the top of the stack. The}
{stacked value must be correctly disposed of later of course.}
{Conventional Three-Byte instruction.}
{1 =  The opcode}
{2 = The level in the display vector of the base address}
{3..WordSize = The offset from that base address of the actual value}
var Level,Offset,Inx: integer;
begin
Level:=I2LProgram.GetProgramByte;
Offset:=I2LProgram.GetProgramWord;
Inx:=DisplayVector[Level]+Offset;
I2LProgram.SetDataWord(Inx,I2LProgram.GetDataWord(Inx)+1);
PushStack(I2LProgram.GetDataWord(Inx));
end;



procedure ORDO;
{ $1A,1, OR}
begin
PushStack(PopStack or PopStack);
end;


procedure ANDDO;
{ $1B,1, AND}
begin
PushStack(PopStack and PopStack);
end;


procedure NOTDO;
{ $1C,1, 1'S COMPLEMENT}
begin
PushStack(not PopStack);
end;


procedure EORDO;
{ $1D,1, EOR}
begin
PushStack(PopStack xor PopStack);
end;


procedure DOUBL;
{ $1E,1, NOS + TOS * IntBytes (DBA)}
{ Routine to form TOS * IntBytes + NOS onto TOS (DBA). One-byte Instruction.}
begin
PushStack(PopStack * IntBytes + PopStack);
end;


procedure DEFSAV;
{ $1F,1, INDIRECT SAVE (STD)}
{STORE TOS INTO ADDRESS IN NOS AND POP BOTH (STD). ONE-BYTE}
{INSTRUCTION.}
var Tos: TI2LWord;
begin
Tos:=PopStack;
I2LProgram.SetDataWord(PopStack,Tos);
end;



procedure DEFER;
{ $20,1, INDIRECT GET (DBX)}
{This routine forms TOS*IntBytes+NOS and then uses that value as an}
{Indirect Address of a word to push onto TOS (DBX).}
{One-Byte Instruction.}
var Inx: TI2LWord;
begin
Inx:=PopStack * IntBytes + PopStack;
PushStack(I2LProgram.GetDataWord(Inx));
end;



procedure ADDRDO;
{ $21,4, ADDRESS OF VARIABLE}
{Load the Address of a Variable onto the Stack.}
{Conventional Three-Byte Instruction.}
{1 =  The opcode}
{2 = The level in the display vector of the base address}
{3 = The offset from that base address of the actual value}
var Level,Offset: integer;
begin
Level:=I2LProgram.GetProgramByte;
Offset:=I2LProgram.GetProgramWord;
PushStack(DisplayVector[Level]+Offset);
end;



procedure LDIDO;
{ $22,1, INDIRECT GET}
{Routine to replace TOS by the Integer it points to.}
{One-Byte Instruction.}
begin
PushStack(I2LProgram.GetDataWord(PopStack));
end;



procedure LDADO;
{ $23,3, ABSOLUTE GET}
{Instruction to get an Integer from an Absolute Address.}
{Three-Byte Instruction:}
{1 = OPCODE}
{2 = ADDRESS}
begin
PushStack(I2LProgram.GetDataWord(I2LProgram.GetProgramWord));
end;



procedure SIMMDO;
{ $24,2, SHORT IMMEDIATE}
{Short Immediate Load of an 8-Bit, Signed value.}
{ Two-Byte Instruction:}
{1 = OPCODE}
{2 = VALUE TO LOAD}
begin
{Use type-cast to force it to be signed}
PushStack(ShortInt(I2LProgram.GetProgramByte));
end;



procedure CAJMP;
{ $25,3, CASE JUMP}
{THIS ROUTINE OPTIMIZES THE CASE STATEMENT A LITTLE BIT.}
{IT POPS TOS, COMPARES TOS TO NOS AND TAKES THE BRANCH}
{ONLY IF NOS WAS NOT EQUAL TO TOS.}
{1 = OPCODE}
{2..WordSize = Jump Target}
begin
if PopStack<>I2LStack[StackPtr-1].Item then I2LProgram.ProgCntr:=I2LProgram.GetProgramWord
else I2LProgram.GetProgramWord;
end;



procedure JSRDO;
{ $26,3, OPTIMIZED PROC CALL}
{OPTIMIZED PROCEDURE CALL. USED ONLY IF NO LOCAL VARAIBLES ARE}
{PRESENT.  SCOPE IS UNCHANGED THUS IT IS EQUIVALENT TO A MACHINE}
{LANGUAGE JSR.}
{1 = OPCODE}
{2..WordSize = Jump Target}
begin
PushSubCall(I2LProgram.ProgCntr+IntBytes);
I2LProgram.ProgCntr:=I2LProgram.GetProgramWord;
end;




procedure RTSDO;
{ $27,1, OPTIMIZED RETURN}
{OPTIMIZED RETURN, TO MATCH THE ABOVE CALL}
begin
I2LProgram.ProgCntr:=PopStack;
end;




procedure DRPDO;
{ $28,1, DROP TOS}
{ROUTINE TO POP TOS & DISCARD IT}
begin
Dec(StackPtr);
end;


procedure EXTDO;
{ $29,3, CALL EXTERNAL SUBROUTINE}
{CALL EXTERNAL SUBROUTINE.  THIS ALLOWS ACCESS TO APPLICATION-}
{SPECIFIC, MACHINE-LANGUAGE SUBROUTINES. THE ABSOLUTE ADDRESS  }
{APPEARS IN THE INSTRUCTION DIRECTLY.  ARGUMENTS, IF ANY, ARE}
{PASSED VIA THE STACK.}
{1 = OPCODE}
{2..WordSize = Call Target}
var Address: TI2LWord;
begin
Address:=I2LProgram.GetProgramWord;
asm
 PUSH	CS;
 Call	Address;
 end;
end;



procedure FLODDO;
{ $2A,4, FLOATING LOAD}
{ROUTINE TO RETRIVE A FLOATING POINT VARIABLE'S VALUE}
{AND PUSH IT ON THE STACK}
{THREE-BYTE OPCODE:}
{1 =  THE OPCODE}
{2 =  THE LEVEL IN THE DISPLAY VECTOR OF THE BASE ADDRESS}
{3..WordSize =  THE OFFSET FROM THAT BASE ADDRESS OF THE ACTUAL VALUE}
var Level,Offset,Inx: integer;
var I: Integer;
begin
Level:=I2LProgram.GetProgramByte;
Offset:=I2LProgram.GetProgramWord;
Inx:=DisplayVector[Level]+Offset;
for I:=0 to IntsPerReal-1 do
	PushStack(I2LProgram.GetDataWord(Inx+I*IntBytes));
end;



procedure FSTODO;
{ $2B,4, FLOATING STORE}
{ROUTINE TO STORE TOP OF STACK (TOS) INTO A VARIABLE}
{1 =  THE OPCODE}
{2 =  THE LEVEL IN THE DISPLAY VECTOR OF THE BASE ADDRESS}
{3..WordSize =  THE OFFSET FROM THAT BASE ADDRESS OF THE ACTUAL VALUE}
var Level,Offset,Inx: integer;
var I: Integer;
begin
Level:=I2LProgram.GetProgramByte;
Offset:=I2LProgram.GetProgramWord;
Inx:=DisplayVector[Level]+Offset;
for I:=IntsPerReal-1 downto 0 do
	I2LProgram.SetDataWord(Inx+I*IntBytes,PopStack);
end;



procedure FIMMDO;
{ $2C,9, FLOATING LOAD IMMED}
{IMMEDIATE LOAD OF A 64-BIT REAL CONSTANT ONTO THE STACK}
{1 =  THE OPCODE}
{2..Real Size = Floating point number}
var I: integer;
begin
for I:=0 to IntsPerReal-1 do
	PushStack(I2LProgram.GetProgramWord);
end;



procedure FADDDO;
{ $2D,1, FLOATING ADDITION}
{TOS:=NOS + TOS}
begin
PushDouble(PopDouble + PopDouble);
HandleFloatErrors;
end;


procedure FSUBDO;
{ $2E,1, FLOATING SUBTRACT}
{TOS:=NOS - TOS}
begin
PushDouble(-PopDouble + PopDouble);
HandleFloatErrors;
end;


procedure FMULDO;
{ $2F,1, FLOATING MULTIPLY}
{TOS:=NOS * TOS}
begin
PushDouble(PopDouble * PopDouble);
HandleFloatErrors;
end;


procedure FDIVDO;
{ $30,1, FLOATING DIVIDE}
{TOS:=NOS / TOS}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
PushDouble(Nos/Tos);
HandleFloatErrors;
end;



procedure FNEGDO;
{ $31,1, FLOATING NEGATE}
{TOS:=-TOS}
begin
PushDouble(-PopDouble);
end;


procedure FEQDO;
{ $32,1, TEST FLOATING FOR =}
{TOS:= NOS = TOS}
begin
if PopDouble = PopDouble then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;


procedure FNEDO;
{ $33,1, TEST FLOATING FOR #}
{TOS:= NOS # TOS}
begin
if PopDouble <> PopDouble then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;



procedure FGEDO;
{ $34,1, TEST FLOATING FOR >=}
{TOS:= NOS >= TOS}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
if Nos >= Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;


procedure FGTDO;
{ $35,1, TEST FLOATING FOR >}
{TOS:= NOS > TOS}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
if Nos > Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;



procedure FLEDO;
{ $36,1, TEST FLOATING FOR <=}
{TOS:= NOS <= TOS}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
if Nos <= Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;



procedure FLTDO;
{ $37,1, TEST FLOATING FOR <}
{TOS:= NOS < TOS}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
if Nos < Tos then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;



procedure TRADO;
{ $38,1, TOS:= RealSize * TOS + NOS}
{Multiply TOS (The Subscript) by "Real Size," and Add}
{ NOS (The base address) to it.}
{ TOS:=TOS * RealSize + NOS}
{ This op is used to store indexed Reals. It is the}
{ Floating Point counterpart to "DBA"}
begin
PushStack(RealSize * PopStack + PopStack);
end;



procedure TRXDO;
{ $39,1, PUSH @(RealSize*TOS+NOS)}
{ MULTIPLY TOS (THE SUBSCRIPT) BY REAL SIZE AND ADD}
{ NOS (THE BASE ADDRESS) TO IT. PUSH THE CONTENTS OF}
{ THIS LOCATION WHICH A POINTER THE NEXT DIMENSION}
{ OF THE ARRAY.}
{ PUSH @(TOS * RealSize + NOS)}
{ THIS IS OP IS ALSO USED TO STORE INDEXED REALS.}
var Inx: integer;
begin
Inx:=RealSize * PopStack + PopStack;
PushStack(I2LProgram.GetDataWord(Inx));
end;



procedure TRIDO;
{ $3A,1, PUSH(8) @ (TOS(4) + NOS(8)) }
{This routine Adds RealSize * TOS + NOS, and uses the sum as an}
{ Absolute Address from which to fetch a value.}
{ PUSH(RealSize) @(RealSize * TOS(IntSize) + NOS(RealSize) )}
{ Note: NOS is integer packed into a Real}
var Inx,I: integer;
begin
Inx:=RealSize * PopStack;
Dec(StackPtr,IntsPerReal-1);
Inx:=Inx+PopStack;
for I:=0 to IntsPerReal-1 do
	PushStack(I2LProgram.GetDataWord(Inx+I*IntBytes));
end;



procedure STTDO;
{ $3B,1, PULL(RealSize) @(NOS(IntSize))}
{STORE THE RealSize BYTES IN TOS INTO THE ABSOLUTE ADDRESS}
{ POINTED TO BY NOS. THIS OPCODE IS USED TO STORE}
{ INDEXED REAL NUMBERS.}
{ PULL(RealSize) @(NOS)}
var D: double;
var Inx: TI2LWord;
var A: PI2LWordArray;
var I: integer;
begin
D:=PopDouble;
Inx:=PopStack;
A:=PI2LWordArray(@D);
for I:=0 to IntsPerReal-1 do
	I2LProgram.SetDataWord(Inx+I*IntBytes,A[I]);
end;



procedure FADDRDO;
{ $3C,4, ADDRESS OF VARIABLE}
{Load the Address of a Variable onto the Stack.}
{And pad it with zeros so it fits into Real-sized variable}
{1 =  The opcode}
{2 = The level in the display vector of the base address}
{3..WordSize = The offset from that base address of the actual value}
var Level,Offset: integer;
var I: integer;
begin
Level:=I2LProgram.GetProgramByte;
Offset:=I2LProgram.GetProgramWord;
PushStack(DisplayVector[Level]+Offset);
for I:=0 to IntsPerReal-2 do PushStack(0);
end;




procedure ASRDO;
{ $3D,1, ARITHMETIC SHIFT RIGHT}
{ NOS ->> TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
PushStack(DoASR(Nos,Tos));
end;



procedure LSLDO;
{ $3E,1, SHIFT LEFT}
{ NOS << TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
PushStack(Nos shl Tos);
end;



procedure LSRDO;
{ $3F,1, SHIFT RIGHT}
{ NOS << TOS}
var Tos,Nos: TI2LWord;
begin
Tos:=PopStack;
Nos:=PopStack;
PushStack((Nos and WordMask) shr Tos);
end;

{------------ Opcodes for 'Segment' Type Varibables---------------------------}
{------------ Byte Segment variables routines---------------------------------}

procedure LDSBDO;
{ $41,1, LOAD SEGMENT BYTE}
{LOAD A BYTE FROM A SEGMENT VARIABLE}
{TOS=OFFSET, NOS=SEGMENT}
var Offset,SegInx: cardinal;
begin
Offset:=PopStack;
SegInx:=PopStack;
PushStack(MemManager.SegArray[SegInx,Offset]);
end;


procedure STSBDO;
{ $44,1, Store segment BYTE}
{Store a BYTE from a segment variable}
{VALUE is in TOS, OFFSET is in NOS and SEGMENT is in NOS}
var Value,Offset,SegInx: cardinal;
begin
Value:=PopStack;
Offset:=PopStack;
SegInx:=PopStack;
MemManager.SegArray[SegInx,Offset]:=Value;
end;


{------------ Integer Segment variables routines---------------------------------}


procedure LDSIDO;
{ $40,1, Load segment INTEGER}
{ Load an INTEGER from a Segment variable}
{TOS=OFFSET, NOS=SEGMENT}
var Offset,SegInx: cardinal;
var PW: ^TI2LWord;
begin
Offset:=PopStack * IntBytes;
SegInx:=PopStack;
PW:=@MemManager.SegArray[SegInx,Offset];
PushStack(PW^);
end;



procedure STSIDO;
{ $43,1, Store segment INTEGER}
{Store an INTEGER from a segment variable}
{VALUE is in TOS, OFFSET is in NOS and SEGMENT is in NOS}
var Value,Offset,SegInx: cardinal;
var PW: ^TI2lWord;
begin
Value:=PopStack;
Offset:=PopStack * intBytes;
SegInx:=PopStack;
PW:=@MemManager.SegArray[SegInx,Offset];
PW^:=Value;
end;

{------------ Short Segment variables routines---------------------------------}


procedure LSHORT;
{ $46,1, Load SHORT SEGMENT REAL}
{ Routine to load a SHORT REAL from SEGMENT type VARIABLE}
{Short is converted to double and placed on the stack}
{Offset is in AX (TOS) and Segment is in NOS}
var Offset,SegInx: cardinal;
var PS: ^single;
begin
Offset:=PopStack * SingleSize;
SegInx:=PopStack;
PS:=@MemManager.SegArray[SegInx,Offset];
PushDouble(PS^);
end;



procedure SSHORT;
{ $47,1, Store SHORT SEGMENT REAL}
{Routine to store a SHORT REAL into a SEGMENT type variable}
{Value is in TOS, OFFSET is in NOS and SEGMENT is in NOS}
var Value: double;
var Offset,SegInx: cardinal;
var PS: ^single;
begin
Value:=PopDouble;
Offset:=PopStack * SingleSize;
SegInx:=PopStack;
PS:=@MemManager.SegArray[SegInx,Offset];
PS^:=Value;
end;

{------------ Real Segment variables routines---------------------------------}


procedure LDSRDO;
{ $42,1, LOAD SEGMENT REAL}
{ROUTINE TO LOAD A REAL FROM SEGMENT TYPE VARIABLE}
{OFFSET IS IN AX (TOS) AND SEGMENT IS IN NOS}
var Offset,SegInx: cardinal;
var PD: ^double;
begin
Offset:=PopStack * RealSize;
SegInx:=PopStack;
PD:=@MemManager.SegArray[SegInx,Offset];
PushDouble(PD^);
end;



procedure STSRDO;
{ $45,1, Store segment REAL}
{Routine to store a REAL into a segment type variable}
{VALUE is in TOS, OFFSET is in NOS and SEGMENT is in NOS}
var Value: double;
var Offset,SegInx: cardinal;
var PD: ^double;
begin
Value:=PopDouble;
Offset:=PopStack * RealSize;
SegInx:=PopStack;
PD:=@MemManager.SegArray[SegInx,Offset];
PD^:=Value;
end;


{------------- Array Reserve Subroutines --------------------------------------}


var DimCnt: TI2lWord;




procedure GenArray(BaseAdd,SizeOfType,DimInx: TI2LWord; DimArray: array of TI2LWord);
var ElmCnt,ResSize,I: integer;
var ResBase: integer;
begin
ResBase:=I2LProgram.HeapPtr;
I2LProgram.SetDataWord(BaseAdd,I2LProgram.HeapPtr);

ResSize:=DimArray[DimInx];
ElmCnt:=ResSize;
{char arrays are stored in integers, except for last dimension}
if (SizeOfType = 1) and
   (DimInx<High(DimArray)) then ResSize:=ResSize * IntBytes
else ResSize:=ResSize * SizeOfType;

ReserveMemory(ResSize);
if DimInx>=High(DimArray) then exit;


for I:=0 to ElmCnt-1 do
	begin
	{char arrays are stored in integers}
	if SizeOfType = 1 then BaseAdd:=ResBase + (I*IntBytes)
	else BaseAdd:=ResBase + (I*SizeOftype);
	GenArray(BaseAdd,SizeOfType,DimInx+1,DimArray);
	end;
end;

{
Frog -----> Frog(0) ----> Frog(0,0) - Frog(0,1) - Frog(0,2)
            Frog(1) ----> Frog(1,0) - Frog(1,1) - Frog(1,2)
            Frog(2) ----> Frog(2,0) - Frog(2,1) - Frog(2,2)
            Frog(3) ----> Frog(3,0) - Frog(3,1) - Frog(3,2)

Example: addresses for 5 x 5 integer array

A(0): 0  ---> A(0,0):  20  A(0,1):  24  A(0,2):  28  A(0,3):  32  A(0,4):  36
A(1): 4  ---> A(1,0):  40  A(1,1):  44  A(1,2):  48  A(1,3):  52  A(1,4):  56
A(2): 8  ---> A(2,0):  60  A(2,1):  64  A(2,2):  68  A(2,3):  72  A(2,4):  76
A(3): 12 ---> A(3,0):  80  A(3,1):  84  A(3,2):  88  A(3,3):  92  A(3,4):  96
A(4): 16 ---> A(4,0): 100  A(4,1): 104  A(4,2): 108  A(4,3): 112  A(4,4): 116

or if done manually would look like this:

A   :=Reserve(5 * IntSize);
A(0):=Reserve(5 * IntSize);
A(1):=Reserve(5 * IntSize);
A(2):=Reserve(5 * IntSize);
A(3):=Reserve(5 * IntSize);
A(4):=Reserve(5 * IntSize);



}

procedure MKARRAY;
{ $48,1, RESERVE MULTIDIMENSIONAL ARRAY}
{routine To Reserve A Multi-dimensional Array}
{ For Example: Mkarray((3, 5, 7, 11), 4, 2, Addr ArrayName);}
{ The Stack Contains:}
{ Tos = The Address of The Pointer To The Array (a Pointer To A Pointer)}
{ Nos = The Number Of Bytes In Each Element Of The Array (1=char, 2/4=int, 8=real, n=Struct)}
{ Nos = The Number Of Dimensions (4)}
{ Nos = 4th Dimension item Count (11)}
{ Nos = 3rd Dimension item Count (7)}
{ Nos = 2nd Dimension item Count (5)}
{ Nos = 1st Dimension item Count (3)}
{    (note: This Routine Sets The Pointer To The Space It Reserves)}
var BaseAdd: TI2lWord;
var SizeOfType: TI2lWord;
var DimItemCnt: array of TI2LWord;
var I: integer;
begin
BaseAdd:=PopStack;
SizeOfType:=PopStack;
DimCnt:=PopStack;
SetLength(DimItemCnt,DimCnt);
for I:=High(DimItemCnt) downto 0 do DimItemCnt[I]:=PopStack;
GenArray(BaseAdd,SizeOfType,0,DimItemCnt);
end;


{------------------------------------------------------------------------------}


procedure FORDDO;
{ $49,3, "FOR" Loop Control for Downto}
{This routine handles the test and branch of the 'for' 'downto' loop.}
{The stack is assumed to contain the limit (NOS) and a copy of}
{the loop control variable (TOS). THey are compared and if the}
{loop variable is less than the limit, this opcode's address}
{is branched to, and the Stack is cleaned up (the 'for' loop is}
{done). otherwise, the PC is advanced to the next opcode, and}
{the limit value is left on the Stack (the 'for' loop}
{continues). The code for this instruction is 3 bytes long:}
{1:	The Opcode}
{2..WordSize = Branch Address}
begin
if PopStack<I2LStack[StackPtr-1].Item then
	begin
	{Skip past for-block and clean stack}
	I2LProgram.ProgCntr:=I2LProgram.GetProgramWord;
	PopStack;
	end
else
	begin
	{Skip the arguments}
	Inc(I2LProgram.ProgCntr,IntBytes);
	end;
end;





procedure DECDO;
{ $4A,4, Decrement AND PUSH}
{ Instruction to Increment a variable's value.}
{It leaves a copy of the variable on the top of the stack. The}
{stacked value must be correctly disposed of later of course.}
{COnventional Three-Byte instruction.}
{1 =  The opcode}
{2 = The level in the display vector of the base address}
{3..WordSize = The offset from that base address of the actual value}
var Level,Offset,Inx: integer;
begin
Level:=I2LProgram.GetProgramByte;
Offset:=I2LProgram.GetProgramWord;
Inx:=DisplayVector[Level]+Offset;
I2LProgram.SetDataWord(Inx,I2LProgram.GetDataWord(Inx)-1);
PushStack(I2LProgram.GetDataWord(Inx));
end;



{------------------------------------------------------------------------------}


procedure ABSDO;
{ $50,1, Integer Absolute Value}
{TOS:= ABS(TOS)}
begin
PushStack(abs(PopStack));
end;




procedure REMDO;
{ $51,1, Get Remaineder}
{TOS:= Remaineder}
begin
PopStack;
PushStack(Remainder);
end;



procedure SWAPDO;
{ $52,1, Swap low and high bytes}
{TOS:= Swap(TOS)}
begin
PushStack(SwapByteArgs(PopStack));
end;



procedure EXTENDDO;
{ $53,1, Sign extend Low part of TOS}
{TOS:= -TOS}
begin
PushStack(ShortInt(Lo(PopStack)));
end;



procedure SQRTDO;
{ $54,1, Integer Square Root}
{TOS:= -TOS}
begin
PushStack(round(sqrt(PopStack)));
end;

procedure SQRDO;
{ $55,1, integer Square }
{TOS:= -TOS}
begin
PushStack(sqr(PopStack));
end;



procedure FIXDO;
{ $56,1, Convert TOS to Integer}
{TOS:=FIX(TOS)}
begin
PushStack(round(PopDouble));
end;


procedure FLOATDO;
{ $57,1, Convert TOS to Real}
{TOS:=SQR(TOS)}
begin
PushDouble(PopStack);
end;


procedure FABSDO;
{ $58,1, Floating Absolute Value}
{TOS:=ABS(TOS)}
begin
PushDouble(abs(PopDouble));
end;


procedure FSQRTDO;
{ $59,1, Square Root}
{TOS:=SQRT(TOS)}
begin
PushDouble(sqrt(PopDouble));
end;


procedure FSQRDO;
{ $5A,1, Square TOS}
{TOS:=SQR(TOS)}
begin
PushDouble(sqr(PopDouble));
end;



procedure STRMODDO;
{ $60,2, Set String Mode}
{Routine to set the string mode flag}
{If the argument is non-zero, strings are MSB-Terminated}
{If the argument is zero, strings are null terminated}
{Two-byte instruction:}
{1 = opcode}
{2 = Mode Flag}
begin
UseMSBStrings:=I2LProgram.GetProgramByte<>0;
end;

{------------------ Structure opcodes -----------------------------------------}



{
Example Struct access

type TVector = record
 real X,Y,Z;
 end

TVector V(0);

V(2).Z

V -----> V(0) = 0 * Struct Size (24)
         V(1) = 1 * Struct Size (24)
         V(2) = 2 * Struct Size (24)
         V(2).X = (2 * Struct Size) + X-Offset (0)
         V(2).Y = (2 * Struct Size) + Y-Offset (8)
         V(2).Z = (2 * Struct Size) + Y-Offset (16)
         V(Inx).X = (Inx * Struct Size) + X-Offset (0)
         V(Inx).Y = (Inx * Struct Size) + Y-Offset (8)
         V(Inx).Z = (Inx * Struct Size) + Y-Offset (16)

Base + Index * Size + Offset
}

var StructSize: integer;	{Saved Global Size of current struct }



function GetStructAddress(Indirect: boolean): integer;
{Get base address of struct from stack and return the final address }
{Routine gets TOS + NOS * NOS + NOS, which is: Offset + Index * Size + Base address}
var Offset,Index,Base: TI2LWord;
begin
Offset:=PopStack;
Index:=PopStack;
StructSize:=PopStack;
Base:=PopStack;
if Indirect then Result:=I2LProgram.GetDataWord(Base + Offset)+StructSize * Index
else Result:=Offset + (StructSize * Index) + Base;
end;



procedure LoadStructValue(VarType: TDataType; Indirect: boolean);
{Loads struct where last item is an array array item down to specified component}
{Routine gets TOS + NOS * NOS + NOS, which is: Offset + Index * Size + Base address}
{As an address to push a byte, integer, or real on the stack}
var I: integer;
var Add: TI2LWord;
var Offset,Index,Base: TI2LWord;
begin
Add:=GetStructAddress(Indirect);
case VarType of
 dtChar: PushStack(I2LProgram.DataByte[Add]);
 dtInteger: PushStack(I2LProgram.GetDataWord(Add));
 dtReal: for I:=0 to IntsPerReal-1 do PushStack(I2LProgram.GetDataWord(Add+I*IntBytes));
 end;
end;

{------------- Struct Loads -------------------------------------}

procedure LSXDO;
{ $61,1, Load Structure Indexed (LSX)}
{One-Byte INSTRUCTION.}
{TOS = ItemSize = the size of the subcomponent of the struct}
{NOS = Offset, NOS = Index, NOS = RecordSize, NOS = Base}
{ ItemSize = the size of the subcomponent of the struct}
{ RecordSize = the size of the whole struct }
{ Target address = Offset + (Index * RecordSize) + Base address}
{ Struct at Target Address pushed on stack}
var I: integer;
var Offset,Index,RecordSize,ItemSize,Base,Inx: TI2LWord;
begin
ItemSize:=PopStack;
Offset:=PopStack;
Index:=PopStack;
RecordSize:=PopStack;
Base:=PopStack;
Inx:=Offset + (Index * RecordSize) + Base;
for I:=0 to (ItemSize div IntBytes)-1 do
	PushStack(I2LProgram.GetDataWord(Inx+I*IntBytes));
end;


procedure LIRSDO;
{ $62,2, Load Indexed Real Struct (LIRS) }
{Two byte op code, if second byte is non-zero, load indirect}
{Routine gets TOS + NOS * NOS + NOS, which is: Offset + Size * Index + Base address}
{As an address to push all words of real on stack}
var Indirect: boolean;
begin
Indirect:=I2LProgram.GetProgramByte<>0;
LoadStructValue(dtReal,Indirect);
end;


procedure LIISDO;
{ $63,1, Load Indexed Integer Struct (LIIS) }
{Two byte op code, if second byte is non-zero, load indirect}
{Routine gets TOS + NOS * NOS + NOS, which is: Offset + Size * Index + Base address}
{As an address to push all words of real on stack}
var Indirect: boolean;
begin
Indirect:=I2LProgram.GetProgramByte<>0;
LoadStructValue(dtInteger,Indirect);
end;


procedure LIBSDO;
{ $64,1, Load Indexed Byte Struct (LIBS) }
{Routine gets TOS + NOS * NOS + NOS, which is: Offset + Size * Index + Base address}
{As an address to push all words of real on stack}
begin
LoadStructValue(dtChar,True);
end;

{------------- Struct Stores -------------------------------------}

procedure SIRSDO;
{ $65,1, Store Indexed Real Struct (SIRS)}
{ Pop(RealSize) and store at NOS }
var D: double;
var Inx: TI2LWord;
var A: PI2LWordArray;
var I: integer;
begin
D:=PopDouble;
Inx:=PopStack;
//Inx:=GetStructAddress(False);
A:=PI2LWordArray(@D);
for I:=0 to IntsPerReal-1 do
	I2LProgram.SetDataWord(Inx+I*IntBytes,A[I]);
end;


procedure SIISDO;
{ $66,1, Store Indexed Integer Struct (SIIS)}
{ Pop(integer) and store at NOS}
var Inx, Value: TI2LWord;
begin
Value:=PopStack;
Inx:=PopStack;
//Inx:=GetStructAddress(False);
I2LProgram.SetDataWord(Inx,Value);
end;


procedure SIBSDO;
{ $67,1, Store Indexed Byte Struct (SIBS)}
{ Pop(Byte) and store at NOS}
var Inx, Value,Add: TI2LWord;
begin
Value:=PopStack;
Inx:=PopStack;
//Inx:=GetStructAddress(True);
I2LProgram.DataByte[Inx]:=Value;
end;



procedure SSXDO;
{ $68,1, Store Structure Indexed (SSX)}
{Routine pulls from stack Size, struct data and Base address}
{One-byte instruction:}
var I: integer;
var Size,Base: TI2LWord;
begin
Size:=PopStack;
{Base is underneath data, so grab it out from under}
Base:=GetStackOffset(Size div IntBytes);
for I:=(Size div IntBytes)-1 downto 0 do
	I2LProgram.SetDataWord(Base+I*IntBytes,PopStack);
{Remove stack Base from stack}
PopStack;
end;



procedure PISADO;
{ $69,1, Push Indexed Struct Address }
{ Push TOS + NOS * NOS + NOS, which is: Offset + (Index * Size) + Base address}
{ Used to get address of an Indexed Struct}
var Index: integer;
begin
Index:=GetStructAddress(False);
PushStack(Index);
end;

procedure PIIADO;
{ $6C,1, Push Indexed Indirect Struct Address }
{ Push TOS + NOS * NOS + NOS, which is: Offset + (Index * Size) + Base address}
{ Used to get address of an Indexed Struct}
var Index: integer;
begin
Index:=GetStructAddress(True);
PushStack(Index);
end;

procedure PINSDO;
{ $6A,1, Push indirect struct}
{Takes TOS * NOS + NOS, which is: Index * Size + Base}
{And uses it as an address, whose content is pushed on the stack}
var Inx: TI2LWord;
var Size, Base, Index: integer;
begin
Index:=PopStack;
Size:=PopStack;
Base:=PopStack;
Inx:=Size * Index + Base;
PushStack(I2LProgram.GetDataWord(Inx));
end;



procedure SSHPDO;
{ $6B,1, Store Structure On top of Heap (SSHP)}
{Routine pulls struct-size and data from stack and stores it at top heap}
{Store heap address at global zero so data can be retrieved by caller}
{Two-byte instruction:}
{1 = opcode}
var I: integer;
var Size,Base: TI2LWord;
begin
Size:=PopStack;
Base:=I2LProgram.HeapPtr;
{Store struct address at global-0}
I2LProgram.SetDataWord(DisplayVector[0],Base);
for I:=(Size div IntBytes)-1 downto 0 do
	I2LProgram.SetDataWord(Base+I*IntBytes,PopStack);
end;


{------------------------------------------------------------------------------}

function InterpretLoop(OneStep: boolean): boolean;
begin
Clear8087Status;
StopProgram:=OneStep;
Result:=False;
repeat
	begin
	LastPC:=I2LProgram.ProgCntr;
	OpCode:=I2LProgram.ProgData.Code[I2LProgram.ProgCntr];
	Inc(I2LProgram.ProgCntr);
	if (OpCode and $80)<>0 then
		begin
		{Global load - don't handle in subroutine for speed}
		{ Level = DisplayVector[0]  and  (OpCode and $7F) = offset}
		PushStack(I2LProgram.GetDataWord(DisplayVector[0] + OpCode and $7F));
		end
	else case Opcode of
	 $00: break;		{ $00,1, EXIT INTERPRETER}
	 $01: LODDO;		{ $01,4, LOAD A VARIABLE}
	 $02: LDXDO;		{ $02,1, INDEXED LOAD A BYTE}
	 $03: STODO;		{ $03,4, STORE INTO A VARIABLE}
	 $04: STXDO;		{ $04,1, INDEXED STORE TO A BYTE}
	 $05: CALDO;		{ $05,4, CALL AN I2L PROCEDURE}
	 $06: RETDO;		{ $06,1, RETURN FROM I2L PROCEDURE}
	 $07: JMPDO;		{ $07,3, JUMP TO I2L CODE}
	 $08: JPCDO;		{ $08,3, JUMP IF FALSE}
	 $09: HPIDO;		{ $09,3, INCREMENT HEAP PTR BY ARG}
	 $0A: ARGDO;		{ $0A,3, GET PROC ARGUMENTS}
	 $0B: IMMDO;		{ $0B,3, IMMEDIATE LOAD OF ARG}
	 $0C: CMLDO;		{ $0C,2, CALL AN INTRINSIC}
	 $0D: ADDDO;		{ $0D,1, ADD}
	 $0E: SUBDO;		{ $0E,1, SUBTRACT}
	 $0F: MULDO;		{ $0F,1, MULTIPLY}
	 $10: DIVDO;		{ $10,1, DIVIDE}
	 $11: NEGDO;		{ $11,1, UNARY MINUS}
	 $12: EQDO;		{ $12,1, TEST FOR EQUAL}
	 $13: NEDO;		{ $13,1, TEST FOR NOT EQUAL}
	 $14: GEDO;		{ $14,1, TEST FOR >=}
	 $15: GTDO;		{ $15,1, TEST FOR >}
	 $16: LEDO;		{ $16,1, TEST FOR <=}
	 $17: LTDO;		{ $17,1, TEST FOR <}
	 $18: FORDO;		{ $18,3, 'FOR' LOOP CONTROL}
	 $19: INCDO;		{ $19,4, INCREMENT AND PUSH}
	 $1A: ORDO;		{ $1A,1, OR}
	 $1B: ANDDO;		{ $1B,1, AND}
	 $1C: NOTDO;		{ $1C,1, 1'S COMPLEMENT}
	 $1D: EORDO;		{ $1D,1, EOR}
	 $1E: DOUBL;		{ $1E,1, NOS+TOS*2 (DBA)}
	 $1F: DEFSAV;		{ $1F,1, INDIRECT SAVE (STD)}
	 $20: DEFER;		{ $20,1, INDIRECT GET (DBX)}
	 $21: ADDRDO;		{ $21,4, ADDRESS OF VARIABLE}
	 $22: LDIDO;		{ $22,1, INDIRECT GET}
	 $23: LDADO;		{ $23,3, ABSOLUTE GET}
	 $24: SIMMDO;		{ $24,2, SHORT IMMEDIATE}
	 $25: CAJMP;		{ $25,3, CASE JUMP}
	 $26: JSRDO;		{ $26,3, OPTIMIZED PROC CALL}
	 $27: RTSDO;		{ $27,1, OPTIMIZED RETURN}
	 $28: DRPDO;		{ $28,1, DROP TOS}
	 $29: EXTDO;		{ $29,3, CALL EXTERNAL SUBROUTINE}
	 $2A: FLODDO;		{ $2A,4, FLOATING LOAD}
	 $2B: FSTODO;		{ $2B,4, FLOATING STORE}
	 $2C: FIMMDO;		{ $2C,9, FLOATING LOAD IMMED}
	 $2D: FADDDO;		{ $2D,1, FLOATING ADDITION}
	 $2E: FSUBDO;		{ $2E,1, FLOATING SUBTRACT}
	 $2F: FMULDO;		{ $2F,1, FLOATING MULTIPLY}
	 $30: FDIVDO;		{ $30,1, FLOATING DIVIDE}
	 $31: FNEGDO;		{ $31,1, FLOATING NEGATE}
	 $32: FEQDO;		{ $32,1, TEST FLOATING FOR =}
	 $33: FNEDO;		{ $33,1, TEST FLOATING FOR #}
	 $34: FGEDO;		{ $34,1, TEST FLOATING FOR >=}
	 $35: FGTDO;		{ $35,1, TEST FLOATING FOR >}
	 $36: FLEDO;		{ $36,1, TEST FLOATING FOR <=}
	 $37: FLTDO;		{ $37,1, TEST FLOATING FOR <}
	 $38: TRADO;		{ $38,1, TOS:= 5*TOS +NOS}
	 $39: TRXDO;		{ $39,1, PUSH @(5*TOS+NOS)}
	 $3A: TRIDO;		{ $3A,1, PUSH(5)@(TOS(2)+NOS(5))}
	 $3B: STTDO;		{ $3B,1, PULL(5) @(NOS(2))}
	 $3C: FADDRDO;		{ $3C,1, FADDRDO}
	 $3D: ASRDO;		{ $3D,1, Arithmetic Shift Right }
	 $3E: LSLDO;		{ $3E,1, SHIFT LEFT}
	 $3F: LSRDO;		{ $3F,1, SHIFT RIGHT}
	 $40: LDSIDO;		{ $40,1, LOAD SEGMENT INTEGER}
	 $41: LDSBDO;		{ $41,1, LOAD SEGMENT BYTE}
	 $42: LDSRDO;		{ $42,1, LOAD SEGMENT REAL}
	 $43: STSIDO;		{ $43,1, STORE SEGMENT INTEGER}
	 $44: STSBDO;		{ $44,1, STORE SEGMENT BYTE}
	 $45: STSRDO;		{ $45,1, STORE SEGMENT REAL}
	 $46: LSHORT;		{ $46,1, LOAD SHORT SEGMENT REAL}
	 $47: SSHORT;		{ $47,1, STORE SHORT SEGMENT REAL}
	 $48: MKARRAY;		{ $48,1, RESERVE MULTIDIMENSIONAL ARRAY}
	 $49: FORDDO;		{ $49,3, 'FOR' LOOP CONTROL FOR DOWNTO}
	 $4A: DECDO;		{ $4A,4, Decrement AND PUSH}

	 $4B: OPERR;		{ $4B,1, BAD OPCODE}
	 $4C: OPERR;		{ $4C,1, BAD OPCODE}
	 $4D: OPERR;		{ $4D,1, BAD OPCODE}
	 $4E: OPERR;		{ $4E,1, Reserved for Port In}
	 $4F: OPERR;		{ $4F,1, Reserved for Port Out}

	 $50: ABSDO;		{ $50,1, Integer Absolute Value}
	 $51: REMDO;		{ $51,1, Remainder }
	 $52: SWAPDO;		{ $52,1, Swap Bytes }
	 $53: EXTENDDO;		{ $53,1, Extend Sign }
	 $54: SQRTDO;		{ $54,1, Integer Square Root }
	 $55: SQRDO;		{ $55,1, Integer Square }
	 $56: FIXDO;		{ $56,1, Convert Real to Integer Number }
	 $57: FLOATDO;		{ $57,1, Convert Integer to Real Number }
	 $58: FABSDO;		{ $58,1, Floating Point ABS }
	 $59: FSQRTDO;		{ $59,1, Floating Point Square Root }
	 $5A: FSQRDO;		{ $5A,1, Floating Point Square }

	 $5B: OPERR;		{ $5B,1, BAD OPCODE}
	 $5C: OPERR;		{ $5C,1, BAD OPCODE}
	 $5D: OPERR;		{ $5D,1, BAD OPCODE}
	 $5E: OPERR;		{ $5E,1, BAD OPCODE}
	 $5F: OPERR;		{ $4F,1, BAD OPCODE}
	 $60: STRMODDO;		{ $60,2, Set String Mode }

	 $61: LSXDO;		{ $61,1, Load Struct Indexed (LSX)}
	 $62: LIRSDO;		{ $62,2, Load Indexed Real Struct (LIRS)}
	 $63: LIISDO;		{ $63,2, Load Indexed Integer Struct (LIIS) }
	 $64: LIBSDO;		{ $64,1, Load Indexed Byte Struct (LIBS) }

	 $65: SIRSDO;		{ $65,1, Store Indexed Real Struct (SIRS)}
	 $66: SIISDO;		{ $66,1, Store Indexed Integer Struct (SIIS)}
	 $67: SIBSDO;		{ $67,1, Store Indexed Byte Struct (SIBS)}
	 $68: SSXDO;		{ $68,1, Store Structure Indexed (SSX)}

	 $69: PISADO;           { $69,1, Push Indexed Struct Address (PISA) }
	 $6A: PINSDO;		{ $6A,1, Push indirect struct (PINS)}
	 $6B: SSHPDO;		{ $6B,1, Store Structure On top of Heap (SSHP)}
	 $6C: PIIADO;		{ $6C,1, Push Indexed Indirect Struct Address }
	 $75: begin Result:=True; exit; end;
	 else OPERR;
	 end;
	Inc(LoopCount);
	if (LoopCount mod 1000000)=0 then
		begin
		Application.ProcessMessages;
		{Make sure there are no floating point errors after context switch}
		Clear8087Status;
		end;
	end
until StopProgram;
end;


{------------------------------------------------------------------------------}



procedure InitializeInterpreter(Console: TConsoleWindow);
var I: integer;
begin
UseMSBStrings:=True;
InitializeIntrinsics(Console);
InitializeConsole;
I2LProgram.ProgCntr:=0;
StackPtr:=0;
I2LProgram.HeapPtr:=I2LProgram.HeapLow;
Traps:=$7FFF;
for I:=0 to High(DisplayVector) do DisplayVector[I]:=0;
DisplayVector[0]:=I2LProgram.HeapPtr;
DisplayPtr:=0;
PushStack(DisplayPtr);
PushStack(DisplayVector[DisplayPtr]);
PushStack(I2LProgram.ProgCntr);
Inc(I2LProgram.ProgCntr);
InitalizeRedirection;
end;





procedure FinalizeInterpreter;
begin
FinalizeRedirection;
FinalizeIntrinsics;
I2LProgram.ReleaseMemory;
end;


procedure StartProgram(Console: TConsoleWindow);
begin
InitializeInterpreter(Console);
PauseFlag:=False;
try
while true do
	begin
	InterpretLoop(False);
	{Did we exit with a Pause? If not then just exit;}
	if not PauseFlag then break;
	{Undo abort since we are only pausing}
	StopProgram:=False;
	{Wait for user to un-pause }
	while PauseFlag do
		begin
		Application.ProcessMessages;
		{Exit if user aborts}
		if StopProgram then exit;
		end
	end;
finally FinalizeInterpreter; end;
end;



procedure KillProgram;
begin
StopProgram:=True;
AbortConsoleFunctions;
end;



procedure SetProgramCode(var Code: TI2LDataBlock);
begin
I2LProgram:=Code;
I2LProgram.InitializeHeap(32767,MaxInteger,1024);
end;


procedure Interpreter(var Code: TI2LDataBlock; Console: TConsoleWindow);
begin
SetProgramCode(Code);
StartProgram(Console);
end;



procedure RunProgram(Source: string; Console: TConsoleWindow);
begin
if not LoadI2LCode(Source,ProgramStore.ProgData) then exit;
Console.VertScrollBar.Position:=0;
Console.HorzScrollBar.Position:=0;
Interpreter(ProgramStore,Console);
end;


initialization
MemManager:=TMemoryManager.Create;
finalization
MemManager.Free;
end.
