unit I2LLoader;

interface

uses Windows,SysUtils,DSupport,I2lGlobal,I2LDataTypes;

function LoadI2LCode(Source: string; var Prog: TI2LProgram): boolean;

implementation

{$R+}

uses XPLDataTypes;

{

Load format is normal ASCII interpreted as follows:

1. ; = Relative load address follows (4 chars). The PC is
set to this value plus the Base Load Address.

2. * = Relative address follows (4 chars). This address is
to be relocated by adding its module load address to it.

3. ^ = Relative address follows (4 chars). This address is
to have the current PC value stored into it. (This enables
forward references to be resolved by the loader.)

4. % = No longer handled directly by the loader, this
command is handled by xlink. It is used to indicate that the
following ascii string is a label which is the start of a
'public' procedure. The label is ignored here so that
'public' is ignored unless xlink is used.

5. # = No longer handled directly by the loader, this
command is handled by xlink. It is used to indicate that the
following ascii string is reference to a label.

6. $ =	End of (.I2l) file

After any of the above there may follow 2 characters of hex
for each loaded byte for any number of bytes.

Note that the "Base Load Address" and the "Module Load
Address" are are equivalent except when loading secondary
modules.

The byte order of addresses in the ".I2L" file is high, low; but
in the loaded I2L code the order is always low, high.

}


var ProgramCounter: dword;	{Load program counter}
var AbsolutBase: dword;		{Absolute relocation base (#)}
var ModuleBase: dword;		{Module relocation base (@)}
var LinkPointer: dword;		{Link table pointer ('EPROC')}

var ProgramSize: dword;		{Holds program size}
var PROGLO: byte;



var SourceString: string;
var SourceInx: integer;

{---------------- Input Routines ------------------------------------}

function GetChar: char;
{Read a character from the input file}
begin
Result:=SourceString[SourceInx];
Inc(SourceInx);
end;



function GetHexNibble(var N: byte): boolean;
{Get a single hex digit from the load device.}
{Return false if character is not hex}
var C: char;
begin
C:=GetChar;
Result:=IsHexDigit(C);
if Result then N:=StrToInt('$'+C);
end;


function GetHexByte(var B: byte): boolean;
{Get a byte represented in ASCII hex,}
{Return false if bad character}
var N1,N2: byte;
begin
Result:=False;
if not GetHexNibble(N1) then exit;
N1:=N1 shl 4;
if not GetHexNibble(N2) then exit;
B:=N1 or N2;
Result:=True;
end;



function GetAddress(var W: TI2LWord): boolean;
{Read in a value and return it in "ADRES" and W}
var B: byte;
var I: integer;
begin
Result:=False;
for I:=0 to IntBytes-1 do
	begin
	if not GetHexByte(B) then exit;
	W:=TI2LWord((W shl 8) or B);
	end;
end;



function ReadAddress: TI2LWord;
begin
GetAddress(Result);		{Read an address}
Result:=Result+ModuleBase;	{Reloacte it}
end;

{-------------- Routines to store into code array ----------------------------}


procedure StoreChar(B: byte; var Code: TCodeArray);
{Store byte in B at the Program Counter}
begin
if ProgramCounter<=High(Code) then
	begin
	Code[ProgramCounter]:=B;
	Inc(ProgramCounter);
	end
else
	begin
	SetLength(Code,Length(Code)+1);
	Code[High(Code)]:=B;
	ProgramCounter:=Length(Code);
	end;
end;


procedure StoreWord(W: TI2lWord; var Code: TCodeArray);
{Routine to store W as a word in memory}
var I: integer;
begin
for I:=0 to IntBytes-1 do StoreChar(PByteArray(@W)[I],Code);
end;



procedure NewAddress(var Code: TCodeArray);
{ (;) handle new Program Counter address}
var Old,I: integer;
begin
Old:=Length(Code);
ProgramCounter:=ReadAddress;		{(;) fetch new address}
if ProgramCounter>=Length(Code) then
	begin
	SetLength(Code,ProgramCounter+1);
	for I:=Old to ProgramCounter-1 do Code[0]:=0;
	end;
end;



procedure FixForwardRef(var Code: TCodeArray);
{ (^) Handle fixup}
var Inx: TI2LWord;
var I: integer;
begin
Inx:=ReadAddress;			{(^) Get adress to fix}
for I:=0 to IntBytes-1 do Code[Inx+I]:=PByteArray(@ProgramCounter)[I];
end;


procedure DoRelativeAddress(var Code: TCodeArray);
{ (*) handle relative address}
var T: TI2LWord;
begin
T:=ReadAddress;		{(*) get and relocate the address}
StoreWord(T,Code);		{And store it}
end;


procedure HandelEndOfFile(var Code: TCodeArray);
{ ($) handle end of I2L file}
begin
ProgramSize:=Length(Code);		{Set save size}
end;



procedure EatLabel;
{ (%) ignore 'public' labels (xlink was not used but it might not be necessary)}
begin
while GetChar>=#$20 do;
end;



function DecodeChar(C: char; var Code: TCodeArray): boolean;
{Special character dispatcher}
{Returns true if done}
begin
Result:=True;
case C of
 '%': EatLabel;
 '#': begin EatLabel; StoreWord(TI2LWord(WordMask),Code); end;
 ';': NewAddress(Code);
 '^': FixForwardRef(Code);
 '*': DoRelativeAddress(Code);
 '$',EOF_Char: begin HandelEndOfFile(Code); exit; end;
 CR_Char,LF_Char: ;
 else raise Exception.Create('Loader: Bad Character: $'+IntToHex(byte(C),2));
 end;
Result:=False;
end;



function LoadI2LCode(Source: string; var Prog: TI2LProgram): boolean;
var C,C2: char;
var B: byte;
begin
Result:=False;
Prog.Loaded:=False;
SetLength(Prog.Code,0);
if Source='' then exit;
ReRunFlag:=false;
CloseOnTerminate:=False;
SourceString:=Source;
SourceInx:=1;
PROGLO:=Length(Prog.Code);	{SET PROGLO}

StoreChar(0,Prog.Code);			{Store an "exit" code as 1ST byte}
AbsolutBase:=ProgramCounter;	{Set relocation base}
ModuleBase:=ProgramCounter;	{Set module base}
LinkPointer:=ProgramCounter;	{Set link point to external table}
Inc(LinkPointer);		{ which begins after jump to main}

C:=GetChar;		{Get first char of file}
DecodeChar(C,Prog.Code);

while true do
	begin
	C:=GetChar;
	if IsHexDigit(C) then
		begin
		C2:=GetChar;
		B:=StrToInt('$'+C+C2);
		StoreChar(B,Prog.Code);
		end
	else if DecodeChar(C,Prog.Code) then break;
	end;
Prog.BaseAddress:=ModuleBase;
Prog.Loaded:=True;
Result:=True;
end;

{$R-}


end.
