unit IDEEditors;

interface

uses Windows,SysUtils,Classes,Controls,ComCtrls,Dialogs, Menus,
     IniFiles, CodeEditor,DSupport;


{Object to make it easy to iterate through include editors on tab pages}

type TIncludeIterator = class(TObject)
 private
  FPageControl: TPageControl;
 protected
 public
  Index: integer;
  constructor Create(PageControl: TPageControl);
  procedure Reset;
  function GetIteration(var Editor: TCodeEditor): boolean;
 end;

{Creates a Tab Sheet with built in code editor}

type TEditorTabsheet = class(TTabSheet)
 private
 protected
  procedure HandleSaveAs(Sender: TObject);
 public
  CodeEditor: TCodeEditor;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
 end;



procedure SetupEditorDialogs(Editor: TCodeEditor);
procedure SetupEditMenus(Editor: TCodeEditor);
function HandleIncludeSave: word;
procedure FinalizeIncludeFiles;

function AddIncludeEditor(ShowPage,UserOpen: boolean; Filename: string): TCodeEditor;
function FindOrCreateEditor(Filename: string): TCodeEditor;
function ShowOrCreateEditor(Filename: string): TCodeEditor;
procedure ShowEditor(Editor: TCodeEditor);
function GetActiveEditor: TCodeEditor;
function SaveByEditor(Editor: TCodeEditor): Word;
function HandleSaveAll: word;


implementation

uses EBXMain, IDEGlobal, IDEConfig;


{ TIncludeIterator }

constructor TIncludeIterator.Create(PageControl: TPageControl);
begin
FPageControl:=PageControl;
Reset;
end;


function TIncludeIterator.GetIteration(var Editor: TCodeEditor): boolean;
{Move to next available Editor tab and return Editor}
{Return false if no more editors availabel}
begin
Result:=False;
while true do
	begin
	Inc(Index);
	if Index>=FPageControl.PageCount then exit;
	if FPageControl.Pages[Index] is TEditorTabsheet then
		begin
		Editor:=TEditorTabsheet(Form1.EditorPageControl.Pages[Index]).CodeEditor;
		Result:=True;
		exit;
		end;
	end;
end;


procedure TIncludeIterator.Reset;
{Reset iteratore to beginning of tab list}
begin
Index:=-1;
end;


{ TEditorTabsheet }

constructor TEditorTabsheet.Create(AOwner: TComponent);
begin
inherited;
CodeEditor:=TCodeEditor.Create(AOwner);
CodeEditor.Parent:=Self;
CodeEditor.Visible:=True;
CodeEditor.Align:=alClient;
CodeEditor.OnSaveAs:=HandleSaveAs;
end;


destructor TEditorTabsheet.Destroy;
begin
CodeEditor.Free;
inherited;
end;


procedure TEditorTabsheet.HandleSaveAs(Sender: TObject);
{Handler to update the tab caption when filename changes}
begin
if Sender is TCodeEditor then
  Caption:=ExtractFilename(TCodeEditor(Sender).Editor.Filename);
end;

{----------------------------------------------------------------------------}


function GetActiveEditor: TCodeEditor;
{Get currently visible/exposed Code Editor}
begin
if Form1.EditorPageControl.ActivePage is TEditorTabsheet then
 Result:=TEditorTabsheet(Form1.EditorPageControl.ActivePage).CodeEditor
else Result:=Form1.SourceEdit;
end;



procedure SetupEditorDialogs(Editor: TCodeEditor);
{Setup file dialogs internal to a Code Editor}
begin
Editor.OpenDialog.Filter:='xpl Source|*.xpl|Any File (*.*)|*.*';
Editor.OpenDialog.InitialDir:='.';
Editor.OpenDialog.DefaultExt:='xpl';

Editor.SaveDialog.Filter:='xpl Source|*.xpl|Any File (*.*)|*.*';
Editor.SaveDialog.InitialDir:='.';
Editor.SaveDialog.DefaultExt:='xpl';

Editor.SaveDialog.Filename:='Test.xpl';
end;



procedure SetupEditMenus(Editor: TCodeEditor);
{Set pop menu internal to Code Editor}
begin
with Form1 do
 begin
 {Insert at beginning of menu - because each is insert at zero, they will appear in reverse order}
 Editor.InsertMenuItem(0,Editor.EditorMenu,'-',nil);
 Editor.InsertMenuItem(0,Editor.EditorMenu,'Open File At Cursor',FileAtCursorClick);
 Editor.InsertMenuItem(0,Editor.EditorMenu,'Find Declaration',FindDeclaration);
 {Insert at end}
 Editor.AddMenuItem(Editor.EditorMenu,'-',nil);
 Editor.AddMenuItem(Editor.EditorMenu,'Lookup Intrinsic',LookupKeyword,VK_F1);
 Editor.AddMenuItem(Editor.EditorMenu,'Insert Intrinsic Prototype',ShowIntrinsicList1Click,ShortCut(VK_F3,[ssCtrl]));
 Editor.AddMenuItem(Editor.EditorMenu,'-',nil);
 Editor.AddMenuItem(Editor.EditorMenu,'File Properties',FileProperties1Click);
 Editor.AddMenuItem(Editor.EditorMenu,'Insert Intrinsics Declaration',InsertInstrinicDeclaration1Click);
 end;
end;



procedure GetEditorDefaults(Editor: TCodeEditor);
{Set specified Code Editor to user defaults from Ini-File}
var XPLIDEIni: TIniFile;
begin
XPLIDEIni:=TIniFile.Create(IDEIniFilename);
with XPLIDEIni do
 begin
 ReadEditorDefaults(XPLIDEIni,Editor.Editor);
 end;
XPLIDEIni.Free;
end;



function SaveByEditor(Editor: TCodeEditor): Word;
{Save if any changes have been in specified code editor}
{Returns mrCancel if user cancels operation}
var S: string;
begin
Result:=mrOK;
if Editor.Editor.IsModified then
	begin
	ShowEditor(Editor);
	S:=ExtractFilename(Editor.Editor.Filename)+' has Changed'+CRLF_Char;
	S:=S+'Save changes?';
	Result:=MessageDlg(S, mtConfirmation, [mbYes, mbNo, mbCancel], 0);
	if Result=mrYes then Editor.SaveHandler(Form1);
	end
end;



function HandleIncludeSave: word;
{Check all include files to see if they have been modified}
{And give user opertunity to save them if they have changed}
var Iterator: TIncludeIterator;
var Editor: TCodeEditor;
begin
Iterator:=TIncludeIterator.Create(Form1.EditorPageControl);
try
Result:=mrOK;
while Iterator.GetIteration(Editor) do
 if SaveByEditor(Editor)=mrCancel then break;
finally Iterator.Free; end;
end;



function HandleSaveAll: word;
{Save all files that have been modified}
begin
if Form1.SourceEdit.Editor.IsModified then
	begin
	Form1.IDEPageControl.ActivePageIndex:=0;
	Form1.EditorPageControl.ActivePageIndex:=0;
	Result:=MessageDlg('Save Source Changes?', mtConfirmation, [mbYes, mbNo, mbCancel], 0);
	if Result=mrYes then Form1.SaveMnuClick(Form1);
	if Result=mrCancel then exit;
	end;
Result:=HandleIncludeSave;
end;




function GetIncludeCount: integer;
{Get number of include editors active}
begin
{Since first tab is the source }
Result:=Form1.EditorPageControl.PageCount-1;
end;




function AddIncludeEditor(ShowPage,UserOpen: boolean; Filename: string): TCodeEditor;
{Add a new editor to the page control, opening specified file if possible}
{ShowPage = switch to the new page so it is visible}
{UserOpen = Ignore specified filename and let user open file}
var TS: TEditorTabsheet;
begin
{Create New Editor page and attach page control}
TS:=TEditorTabsheet.Create(Form1.EditorPageControl);
TS.PageControl:=Form1.EditorPageControl;
TS.CodeEditor.Name:='IncludeEdit'+IntToStr(GetIncludeCount);
{Configure editor}
SetupEditorDialogs(TS.CodeEditor);
SetupEditMenus(TS.CodeEditor);
GetEditorDefaults(TS.CodeEditor);
TS.CodeEditor.IniFile:=IncludeIni;
TS.CodeEditor.Initialize;
TS.CodeEditor.Editor.OnMouseMove:=Form1.IncludeEditMouseMove;
TS.CodeEditor.OnFileChange:=Form1.HandleIncludeFileChange;
{handle file opening}
if UserOpen then
	begin
	TS.CodeEditor.OpenHandler(Form1);
	Filename:=TS.CodeEditor.Editor.Filename;
	end
else if FileExists(Filename) then TS.CodeEditor.OpenFile(Filename);
TS.Caption:=ExtractFilename(Filename);
{Handling showing new page}
if ShowPage then
	begin
	{To prevent a gray box flashing, repaint before it shown}
	Form1.EditorPageControl.DoubleBuffered:=True;
	Form1.EditorPageControl.Repaint;
	Form1.IDEPageControl.ActivePageIndex:=0;
	Form1.EditorPageControl.ActivePage:=TS;
	if TS.CodeEditor.Editor.CanFocus then TS.CodeEditor.Editor.SetFocus;
	end;
Result:=TS.CodeEditor;
end;



function IsFileOpen(Filename: string; var Editor: TCodeEditor): boolean;
{Determine if the specified file is open in one of the include editors}
{Returns true if file is found, also returns the specific editor}
var Iterator: TIncludeIterator;
begin
Iterator:=TIncludeIterator.Create(Form1.EditorPageControl);
try
Result:=True;
while Iterator.GetIteration(Editor) do
 if Editor.CurrentFile=Filename then exit;
Result:=False;
finally Iterator.Free; end;
end;



procedure ShowEditor(Editor: TCodeEditor);
{Force the page controls to display specified editor}
begin
if Editor.Parent is TEditorTabsheet then
	begin
	Form1.IDEPageControl.ActivePageIndex:=0;
	Form1.EditorPageControl.ActivePage:=TEditorTabsheet(Editor.Parent);
	if Editor.CanFocus then Editor.SetFocus;
	end;
end;

function ShowOrCreateEditor(Filename: string): TCodeEditor;
begin
if IsFileOpen(Filename, Result) then ShowEditor(Result)
else Result:=AddIncludeEditor(True,False, Filename);
end;



function FindOrCreateEditor(Filename: string): TCodeEditor;
{Find and display editor that matches filename }
{ if not found, open the specified file}
begin
if Filename=Form1.SourceEdit.Editor.Filename then
	begin
	Result:=Form1.SourceEdit;
	ShowEditor(Result);
	end
else Result:=ShowOrCreateEditor(Filename);
end;



procedure FinalizeIncludeFiles;
{Free all open include editors}
var Iterator: TIncludeIterator;
var Editor: TCodeEditor;
begin
Iterator:=TIncludeIterator.Create(Form1.EditorPageControl);
try
while Iterator.GetIteration(Editor) do Editor.Finalize;
finally Iterator.Free; end;
end;


end.
