unit IDEVariableLib;

interface

uses Types,SysUtils,DSupport,ScrollEditor,IDEExpression, XPLCompiler, XPLStructs;



type TVarStatus = (vsScopeError,vsAccessError,vsValidAddress);

type TVariableType = (vtInteger,vtChar,vtReal,vtIntConst,vtRealConst,vtUserType,vtNone);

type TVariableFormat = (vfNormal,vfInteger,vfReal,vfString);

type TIntegerFormat = (nrDecimal,nrHex,nrOctal,nrBinary);

type TDisplayType = (dsCharString,dsString,dsInteger,dsReal,dsByte);

type TRealFormat = (rfGeneral,rfFixed,rfScientific,rfEngineering);

type TLookupResult = (lrNoIdent,lrNotFound,lrFound);



type TVarFormat = record
 DisplayType: TDisplayType;
 Digits: integer;
 Radix: TIntegerFormat;
 RealFormat: TRealFormat;
 ForcedArray: boolean;
 DisplayDims: TIntegerDynArray;
 end;


type TVariableInfo = record
 VarText: string;
 BaseName: shortString;
 ProcName: shortstring;
 Filename: shortstring;
 BeginPC,EndPC: integer;
 Level,Offset: integer;
 RealConst: double;
 VarType: TVariableType;
 DeclareLine: integer;
 HasIndices: boolean;
 Indices: TIntegerDynArray;
 DefFormat: TVarFormat;
 Dims: PDynWordArray;
 {$IFDEF STRUCTS}
 IsStruct: boolean;
 StructIndex,StructOffset, StructSize: integer;
 StructComp: string;
 {$ENDIF}
 end;



type TWatchInfo = record
 VarText,Filename: string;
 Line: integer;
 CustomFormat: boolean;
 UserFormat: TVarFormat;
 end;

type TWatchArray = array of TWatchInfo;

var WatchArray: TWatchArray;
var TravStart: integer;		{Starting location of traverse}


function GetWatchString(Info: TVariableInfo; Format: TVarFormat; Line: integer): string;
function GetIndicesString(VI: TVariableInfo): string;
function GetVariableAddress(Info: TVariableInfo; var Add: integer): TVarStatus;

function GetDisplayType(VarType: TVariableType; HasIndices: boolean): TDisplayType; overload;
function GetDisplayType(Info: TVariableInfo): TDisplayType; overload;
function GetIdentAndSubsRL(ED: TScrollEditor; P: TPoint): string;
function ParseVariableString(VS,Filename: string; Line: integer; var VI: TVariableInfo): boolean;
procedure GetVariableInfo(PI: PProcItem; SI: PSymbolItem; var VI: TVariableInfo);
function LookupVariable(VS,Filename: string; Line: integer; var VI: TVariableInfo): TLookupResult;
function GetWatchValue(Inx: integer): string;

function GetIntegerRadixStr(I,Digits: integer; Radix: TIntegerFormat): string;
function GetRealStr(D: double; Format: TVarFormat): string;


implementation

uses I2LGlobal, XPLDataTypes, I2LInterp, IDEDebugger, IDEStructDebug;



function GetIndicesString(VI: TVariableInfo): string;
{Get a string that represents the indices - e.g (1,2,5,8) }
var I: integer;
begin
Result:='';
if not VI.HasIndices then exit;
Result:='(';
for I:=0 to High(VI.Indices) do
	begin
	if I<>0 then Result:=Result+',';
	Result:=Result+IntToStr(VI.Indices[I]);
	end;
Result:=Result+')';
end;


function GetIntegerRadixStr(I,Digits: integer; Radix: TIntegerFormat): string;
{Return a string based on the radix of the watch}
begin
case Radix of
 nrDecimal: Result:=IntToStr(I);
 nrHex: Result:='$'+IntToHex(I,Digits);
 nrOctal: Result:=IntToOctal(I)+' (Oct)';
 nrBinary: Result:='%'+IntToBin(I,32,False);
 end;
end;



function GetDisplayType(VarType: TVariableType; HasIndices: boolean): TDisplayType;
{Work out the default way to display the variable}
{Given the way XPL variables can/can't be used}
begin
{char can only be strings if they aren't arrays}
{or are arrays with more than one dimension}
case VarType of
 vtChar:
	begin
        if HasIndices then Result:=dsCharString
	else Result:=dsString;
	end;
 vtInteger: Result:=dsInteger;
 vtReal: Result:=dsReal;
 else Result:=dsInteger;
 end;
end;


function GetDisplayType(Info: TVariableInfo): TDisplayType;
{Work out the default way to display the variable}
{Given the way XPL variables can/can't be used}
begin
Result:=GetDisplayType(Info.VarType, Length(Info.Indices)>0);
end;



function GetRealStr(D: double; Format: TVarFormat): string;
begin
case Format.RealFormat of
 rfGeneral: Result:=FloatToStrF(D,ffGeneral,15,Format.Digits);
 rfFixed: Result:=FloatToStrF(D,ffFixed,15,Format.Digits);
 rfScientific: Result:=FloatToStrF(D,ffExponent,Format.Digits+1,3);
 rfEngineering: Result:=FloatToEngineerStr(D,Format.Digits);
 else Result:=FloatToStrF(D,ffFixed,15,Format.Digits)
 end;
end;



function GetArrayItem(Inx: integer; Add: integer; Info: TVariableInfo; Format: TVarFormat; ErrorStr: string): string; overload;
{Get single array element based on data type}
var Size: integer;
var D: double;
begin
case Format.DisplayType of
 dsCharString:
	begin
	if ProgramStore.RangeCheck(Add+Inx) then Result:='"'+char(ProgramStore.DataByte[Add+Inx])+'"'
	else Result:=ErrorStr;
	end;
 dsString:
	begin
	if Info.VarType=vtReal then Size:=RealSize else Size:=IntBytes;
	if ProgramStore.RangeCheck(Add+Inx*Size) and
	   ProgramStore.RangeCheck(I2LProgram.GetDataWord(Add+Inx*Size)) then
	   	Result:='"'+ProgramStore.GetDataString(I2LProgram.GetDataWord(Add+Inx*Size),UseMSBStrings)+'"'
	else Result:=ErrorStr;
	end;
 dsInteger:
	begin
	if ProgramStore.RangeCheck(Add+Inx*IntBytes) then Result:=GetIntegerRadixStr(ProgramStore.GetDataWord(Add+Inx*IntBytes),IntNibbles,Format.Radix)
	else Result:=ErrorStr;
	end;
 dsReal:
	begin
	if ProgramStore.RangeCheck(Add+Inx*RealSize) then
		begin
		D:=ProgramStore.GetDataDouble(Add+Inx*RealSize);
		Result:=GetRealStr(D,Format);
		end
	else Result:=ErrorStr;
	end;
 dsByte:
	begin
	if ProgramStore.RangeCheck(Add+Inx) then Result:=GetIntegerRadixStr(ProgramStore.DataByte[Add+Inx],1,Format.Radix)
	else Result:=ErrorStr;
	end;
 end;
end;


function GetArrayItem(Inx: integer; Add: integer; Info: TVariableInfo; Format: TVarFormat): string;  overload;
begin
Result:=GetArrayItem(Inx, Add, Info, Format, 'Access Error');
end;



function GetArrayAddress(Info: TVariableInfo; var Add: integer): boolean;
var I,Inx,StepSize,A: integer;
begin
Result:=False;
case Info.VarType of
  vtReal: StepSize:=RealSize;
  vtInteger: StepSize:=IntBytes;
  vtUserType: StepSize:=Info.StructSize;
 else StepSize:=IntBytes;
 end;
for I:=0 to Length(Info.Indices)-2 do
	begin
	Inx:=Info.Indices[I];
	A:=Add+Inx*StepSize;
	if not I2LProgram.RangeCheck(A) then exit;
	Add:=ProgramStore.GetDataWord(A);
	end;
if Length(Info.Indices)<1 then exit;
Inx:=Info.Indices[Length(Info.Indices)-1];
case Info.VarType of
 vtInteger: Add:=Add+Inx*IntBytes;
 vtChar: Add:=Add+Inx;
 vtReal: Add:=Add+Inx*RealSize;
 vtUserType: Add:=Add+Inx*Info.StructSize;
 end;
Result:=True;
end;



function GetVariableAddress(Info: TVariableInfo; var Add: integer): TVarStatus;
begin
{Make sure the variable is in scope}
Result:=vsScopeError;
if (I2LProgram.ProgCntr<=Info.BeginPC) or
   (I2LProgram.ProgCntr>Info.EndPC) then exit;

Add:=DisplayVector[Info.Level]+Info.Offset;
{Add in offset to struct component}
{Make sure the index points to accessable memory}
{Empty variables have random values that point anywhere}
{No need to range check constants}
Result:=vsAccessError;
if not ((Info.VarType=vtIntConst) or (Info.VarType=vtRealConst)) then
 if not I2LProgram.RangeCheck(Add) then exit;

if Info.HasIndices then
	begin
	Add:=I2LProgram.GetDataWord(Add);
	if not GetArrayAddress(Info, Add) then exit;
	end;
if Info.IsStruct then Add:=Add + Info.StructOffset;
Result:=vsValidAddress;
end;



function GetConstants(Info: TVariableInfo; Format: TVarFormat; var S: string): boolean;
begin
Result:=True;
case Info.VarType of
 vtIntConst: S:=GetIntegerRadixStr(Info.Offset,IntNibbles,Format.Radix);
 vtRealConst: S:=GetRealStr(Info.RealConst,Format);
 else Result:=False;
 end;
end;


{------------------------------------------------------------------------------}


function TraverseDims(Info: TVariableInfo; Format: TVarFormat; var Str: string): boolean;
var It: TIntegerDynArray;
var Add: integer;
var OldLen,ItemCount: integer;
const Limit = 1024 * 8;


	procedure StoreText(S: string);
	begin
	if ItemCount>=TravStart then Str:=Str+S;
	end;


	function OutputDims: string;
	begin
	Info.Indices:=It;
	if GetVariableAddress(Info,Add)=vsValidAddress then
		begin
		Result:=GetArrayItem(0,Add,Info,Format,'?');
		end
	else Result:='?';
	end;


	procedure Iterate(Dim: integer);
	var I: integer;
	begin
	It[Dim]:=0;
	StoreText('(');
	for I:=0 to Format.DisplayDims[Dim]-1 do
		begin
		if ItemCount>Limit then exit;
		if Dim<High(Format.DisplayDims) then Iterate(Dim+1)
		else
			begin
			StoreText(OutputDims);
			Inc(ItemCount);
			end;
                if I<(Format.DisplayDims[Dim]-1) then StoreText(',');
		if Length(Str)-OldLen>80 then
			begin
			StoreText(CRLF_Char);
			OldLen:=Length(Str);
			end;
		Inc(It[Dim]);
		end;
	StoreText(')');
	It[Dim]:=0;
	end;


begin
Result:=False;
if not Format.ForcedArray then exit;
ItemCount:=0;
Info.HasIndices:=True;
OldLen:=0;
SetLength(It,Length(Format.DisplayDims));
Iterate(0);
Result:=True;
end;



function GetWatchItemStr(Info: TVariableInfo; Format: TVarFormat; Offset: integer =0): string;
{Returns a string representation of the specified watch item}
var Add: integer;
var Status: TVarStatus;
begin
if GetConstants(Info,Format,Result) then exit;
if TraverseDims(Info,Format,Result) then exit;

Status:=GetVariableAddress(Info,Add);
case Status of
 vsScopeError: begin Result:='Out of Scope'; exit; end;
 vsAccessError: begin Result:='Access Error'; exit; end;
 end;
Result:=GetArrayItem(0,Add+Offset,Info,Format);
end;



function GetIteratedArray(Info: TVariableInfo; Line: integer): string;
{Iterate through all possible items in an n-dimensional array}
{Number of items in each dimension is specified in Info.Dims}
var D,I: integer;
var Iterator: TIntegerDynArray;


	function GetDimStr: string;
	var I: integer;
	begin
	Result:='(';
	for I:=0 to High(Iterator) do
		begin
		if I<>0 then Result:=Result+',';
		Result:=Result+IntToStr(Iterator[I]);
		end;
	Result:=Result+')';
	end;



	procedure ReadArrayItem;
	begin
	Info.HasIndices:=True;
	Info.Indices:=Iterator;
	Result:=Result+GetWatchString(Info,Info.DefFormat,Line);
	Result:=Result+', ';
	end;


	function AllZeros: boolean;
	var I: integer;
	begin
	Result:=False;
	for I:=0 to High(Iterator) do if Iterator[I]<>0 then exit;
	Result:=True;
	end;


begin
Result:='';
SetLength(Iterator,Length(Info.Dims^));
for I:=0 to High(Iterator) do Iterator[I]:=0;
repeat
	begin
	ReadArrayItem;
	{Increment each dimsion, handle overflow to next dim }
	for D:=High(Info.Dims^) downto 0 do
		begin
		Inc(Iterator[D]);
		if Iterator[D]>=Info.Dims^[D] then Iterator[D]:=0
		else break;
		end;
	end
until AllZeros;
end;



function GetWatchString(Info: TVariableInfo; Format: TVarFormat; Line: integer): string;
begin
if Info.IsStruct then
	begin
	Result:=GetStructStr(Info.VarText,Info.Filename,Line);
	exit;
	end;
Result:=Info.BaseName;
if Info.HasIndices then
	begin
	Result:=Result+GetIndicesString(Info);
	Result:=Result+'='+GetWatchItemStr(Info,Format);
	end
else if Length(Info.Dims^)>0 then
	begin
	Result:=GetIteratedArray(Info,Line);
	end
else Result:=Result+'='+GetWatchItemStr(Info,Format);
end;



{------------------------------------------------------------------------------}


procedure GetVariableInfo(PI: PProcItem; SI: PSymbolItem; var VI: TVariableInfo);


	{$IFDEF STRUCTS}
	procedure GetStructInfo(var VI: TVariableInfo);
	var StructType,StructCompIndex: integer;
	begin
	VI.IsStruct:=True;
	VI.StructIndex:=SI.StructIndex;
	VI.StructComp:=GetComponentString(SI.Name,SI.FullName);
	VI.StructOffset:=GetVariableOffset(SI.StructIndex,VI.StructComp);
	VI.StructSize:=GetVariableSize(SI.StructIndex,VI.StructComp);
	GetVariableType(SI.StructIndex,VI.StructComp,StructType,StructCompIndex);
	case StructType of
	 ADDRVAR: VI.VarType:=vtChar;
	 INVAR: VI.VarType:=vtInteger;
	 RLVAR: VI.VarType:=vtReal;
	 USRTYPE: VI.VarType:=vtUserType;
	 end;
	end;
	{$ENDIF}

begin
{$IFDEF STRUCTS}
VI.IsStruct:=False;
{$ENDIF}

VI.BaseName:=SI.Name;
VI.Filename:=SI.SourceInfo.Filename;
VI.ProcName:=PI.ProcName;
VI.BeginPC:=PI.BeginPC;
VI.EndPC:=PI.EndPC;
VI.Level:=SI.Level;
VI.Offset:=SI.Offset;
VI.RealConst:=SI.RealConst;
VI.DeclareLine:=SI.SourceInfo.Line;
VI.DefFormat.ForcedArray:=False;
VI.Dims:=SI.Dims;
case SI.SymType of
 ADDRVAR: VI.VarType:=vtChar;
 INVAR: VI.VarType:=vtInteger;
 RLVAR: VI.VarType:=vtReal;
 RLCON: VI.VarType:=vtRealConst;
 INCON: VI.VarType:=vtIntConst;
 {$IFDEF STRUCTS}
 USRTYPE: GetStructInfo(VI);
 {$ENDIF}
 else VI.VarType:=vtInteger;
 end;
VI.DefFormat.Digits:=5;
VI.DefFormat.Radix:=nrDecimal;
VI.DefFormat.DisplayType:=GetDisplayType(VI);
VI.DefFormat.RealFormat:=rfGeneral;
end;



function GetSubscripts(ED: TScrollEditor; var P: TPoint): string;
var PCnt: integer;
var C: char;
begin
Result:='';
PCnt:=0;
repeat
	begin
	ED.GetCharAfterLoc(P,C);
	Result:=C+Result;
	case C of
	 '(': Dec(PCnt);
	 ')': Inc(PCnt);
	 end;
	if PCnt<=0 then break;
	end
until not ED.DoBackLoc(P);
end;



function GetIdentAndSubsRL(ED: TScrollEditor; P: TPoint): string;
{Get identifier and any subscripts found - Parsing right to left}
{So that it only gets subscripts if ")" is under cursor}
var C: char;
var R: TRect;
begin
Result:='';
while ED.GetCharAfterLoc(P,C) do
	begin
	if C=')' then Result:=GetSubscripts(ED,P)+Result
	else if C='.' then Result:=C+Result
	else if C in ['A'..'Z','a'..'z','0'..'9','_'] then
		begin
		R:=ED.MarkNearestWord(P);
		Result:=ED.GetMarkedText(R)+Result;
		P:=R.TopLeft;
		end
	else break;
	ED.DoBackLoc(P);
	if (P.X=0) and (P.Y=0) then break;
	end;
EatTrailingChars(Result,'.');
end;



function ParseVariableString(VS,Filename: string; Line: integer; var VI: TVariableInfo): boolean;
{Extract subscript arguments, evaluate the expressions to convert them to numeric values}
var ArgStrs: TStringDynArray;
var Sub, Ident: string;
var Inx: integer;


	procedure ExtractArgStrings(SubStr: String; var Args: TStringDynArray);
	{Extract each comma-separated argument}
	var I: integer;
	var S: string;
	begin
	SetLength(Args,0);
	I:=1;
	while true do
		begin
		S:=ExtractToken(SubStr,',',I);
		if S='' then break;
		SetLength(Args,Length(Args)+1);
		Args[High(Args)]:=S;
		end;
	end;


	procedure GetArgValues(ArgStrs: TStringDynArray; var IntArgs: TIntegerDynArray);
	{Evaluate each argument and return an array of integers}
	var I,V: integer;
	begin
	SetLength(IntArgs,Length(ArgStrs));
	for I:=0 to High(ArgStrs) do
		begin
		V:=EvaluateExpression(ArgStrs[I],Filename,Line);
		IntArgs[I]:=V;
		end;
	end;


begin
Result:=False;
if Length(VS)<1 then exit;
VI.VarText:=VS;
VI.Dims:=nil;
SetLength(VI.Indices,0);
Inx:=Pos('(',VS);
if Inx>=1 then Sub:=Copy(VS,Inx,(Length(VS)-Inx)+1) else Sub:='';
Ident:=Copy(VS,1,Length(VS)-Length(Sub));
if Length(Sub)>0 then
	begin
	if Sub[1]='(' then Delete(Sub,1,1);
	if (Length(Sub)>0) then
		begin
		Inx:=Pos(')',Sub);
		if Inx>0 then Delete(Sub,Inx,(Length(Sub)-Inx)+1);
		end;
	ExtractArgStrings(Sub,ArgStrs);
	GetArgValues(ArgStrs, VI.Indices);
	end;
VI.BaseName:=Ident;
VI.HasIndices:=Length(VI.Indices)>0;
Result:=VI.BaseName<>'';
end;



function LookupVariable(VS,Filename: string; Line: integer; var VI: TVariableInfo): TLookupResult;
var PC: integer;
var PI: PProcItem;
var SI: PSymbolItem;
begin
Result:=lrNoIdent;
if not ParseVariableString(VS,Filename,Line,VI) then exit;
Result:=lrNotFound;
PC:=LineNumberToPC(Filename,Line,ProgramStore.ProgData.BaseAddress);
GetSymbolByName(PC,ProgramStore.ProgData.BaseAddress,VI.BaseName,PI,SI);
if PI=nil then exit;
SI.FullName:=VS;
GetVariableInfo(PI, SI, VI);
Result:=lrFound;
end;



function GetWatchValue(Inx: integer): string;
{Get watch item specified by the watch-array index}
var WI: TWatchInfo;
var VI: TVariableInfo;
var R: TLookupResult;
var Format: TVarFormat;
begin
WI:=WatchArray[Inx];
R:=LookupVariable(WI.VarText,WI.Filename,WI.Line,VI);
if WI.CustomFormat then Format:=WI.UserFormat else Format:=VI.DefFormat;
if R=lrFound then Result:=GetWatchString(VI,Format,WI.Line)
else Result:='No Variable';
end;



end.
