unit XPLIntrinValidate;

{Code to handle intrinsic validation }

interface

uses SysUtils,XPLCompiler;



type TIntrinsicInfo = record
 Name: string;
 TypeCheck: boolean;
 ReturnType: char;
 ArgTypes: string;
 end;

var IntrinsicTable: array [0..255] of TIntrinsicInfo;

type TIntrinsicErrors = (ieNone,ieTooMany,ieTooFew,ieMismatch,ieWarning);
type TReturnError = (reNone,reNeedsFunc,reNeedsProc);


procedure SetInstrinsicInfo(Inx: integer; Name: string; TypeCheck: boolean; ReturnType: char; ArgTypes: string);
procedure ClearIntrinsicTable;
function TestIntrinsicArgs(Inx: integer; Found: string; var Msg: string): TIntrinsicErrors;

function TestReturnErrors(Inx: integer; IsFunction: boolean; var Msg: string): TReturnError;


implementation




procedure ClearIntrinsicTable;
var I: integer;
var Info: TIntrinsicInfo;
begin
Info.TypeCheck:=False;
Info.Name:='';
Info.ReturnType:='V';
Info.ArgTypes:='';

for I:=0 to High(IntrinsicTable) do
	begin
        IntrinsicTable[I]:=Info;
	end;
end;


procedure SetInstrinsicInfo(Inx: integer; Name: string; TypeCheck: boolean; ReturnType: char; ArgTypes: string);
begin
if (Inx<Low(IntrinsicTable)) or (Inx>High(IntrinsicTable)) then exit;
IntrinsicTable[Inx].TypeCheck:=TypeCheck;
IntrinsicTable[Inx].Name:=Name;
IntrinsicTable[Inx].TypeCheck:=TypeCheck;
IntrinsicTable[Inx].ReturnType:=ReturnType;
IntrinsicTable[Inx].ArgTypes:=ArgTypes;
end;


function GetArgStr(C: char): string;
begin
case C of
 'I','A': Result:='int';
 'i','a': Result:='@int';
 'R': Result:='real';
 'r': Result:='@real';
 'd': Result:='addr real';
 'V': Result:='void';
 'S': Result:='struct';
 end;
end;


{

Prototype     -    Actual Arguments
------------------------------------
  int     -       int        = Okay
  int     -       @ int      = Okay
  int     -       real       = Error
  int     -       @ real     = Error
  int     -       addr real  = Okay

@ int     -       int        = Warning
@ int     -       @ int      = Okay
@ int     -       real       = Error
@ int     -       @ real     = Error
@ int     -       addr real  = Error

  real    -       int        = Error
  real    -       @ int      = Error
  real    -       real       = Okay
  real    -       @ real     = Error
  real    -       addr real  = Error

@ real    -       int        = Error
@ real    -       @ int      = Error
@ real    -       real       = Warning
@ real    -       @ real     = Okay
@ real    -       addr real  = Error 

addr real is not allowed in prototypes
so it is not shown in the chart.

}


function CompareTypes(Expected,Found: char): TIntrinsicErrors;
begin
case Expected of
 'I':
	begin
	case Found of
	 'I': Result:=ieNone;
	 'i': Result:=ieNone;
	 'R': Result:=ieMismatch;
	 'r': Result:=ieMismatch;
	 'S': Result:=ieMismatch;
	 'd': Result:=ieNone;
	 end;
	end;
 'i':
	begin
	case Found of
	 'I': Result:=ieWarning;
	 'i': Result:=ieNone;
	 'R': Result:=ieMismatch;
	 'r': Result:=ieMismatch;
	 'd': Result:=ieMismatch;
	 'S': Result:=ieMismatch;
	 end;
	end;
 'R':
	begin
	case Found of
	 'I': Result:=ieMismatch;
	 'i': Result:=ieMismatch;
	 'R': Result:=ieNone;
	 'r': Result:=ieMismatch;
	 'd': Result:=ieMismatch;
	 end;
	end;
 'r':
	begin
	case Found of
	 'I': Result:=ieMismatch;
	 'i': Result:=ieMismatch;
	 'R': Result:=ieWarning;
	 'r': Result:=ieNone;
	 'd': Result:=ieMismatch;
	 'S': Result:=ieMismatch;
	 end;
	end;
 end;
end;






function CompareArgs(Expected,Found: string; var Msg: string; var Error: TIntrinsicErrors): boolean;
{Check each argument to make sure they match type and position}
{Return a message about any mismatches}
var I: integer;
begin
Result:=False;
Msg:='';
I:=1;
while I<=Length(Expected) do
	begin
	Error:=CompareTypes(Expected[I],Found[I]);
	if Error<>ieNone then
		begin
		Msg:='Expected: '+GetArgStr(Expected[I]);
		Msg:=Msg+' Found: '+GetArgStr(Found[I]);
		Msg:=Msg+' Pos: '+IntToStr(I);
		exit;
		end;
	Inc(I);
	end;
Result:=True;
end;



function TestIntrinsicArgs(Inx: integer; Found: string; var Msg: string): TIntrinsicErrors;
{Test for argument errors}
var Expected: string;
var L1,L2: integer;
var Error: TIntrinsicErrors;
begin
Expected:=IntrinsicTable[Inx].ArgTypes;
L1:=Length(Found); L2:=Length(Expected);

if not IntrinsicTable[Inx].TypeCheck then Result:=ieNone
else if L1>L2 then Result:=ieTooMany
else if L1<L2 then Result:=ieTooFew
else if not CompareArgs(Expected,Found,Msg,Error) then Result:=Error
else Result:=ieNone;
end;



function TestReturnErrors(Inx: integer; IsFunction: boolean; var Msg: string): TReturnError;
{Test for return errors}
var Expected: char;
begin
Result:=reNone;
if not IntrinsicTable[Inx].TypeCheck then exit;
Expected:=IntrinsicTable[Inx].ReturnType;
if IsFunction then
	begin
	if Expected='V' then Result:=reNeedsFunc;
	end
else if Expected<>'V' then Result:=reNeedsProc;
end;




end.
